#include <boost/bind.hpp>
#include <boost/thread.hpp>

#include <rsc/logging/LoggerFactory.h>

#include <rsb/Factory.h>

#include <cca/Config.h>
#include <cca/Beat.h>
#include <cca/ProcessingStrategy.h>
#include <cca/processing/TimedProcessing.h>

#include "rci/nodes/ResourceNode.h"
#include "rci/nodes/Controller.h"
#include "rci/nodes/debug/Task.h"
#include "rci/nodes/debug/Status.h"

#include "rci/dto/JointAngles.h"
#include "rci/dto/converter/JointAnglesConverter.h"

using namespace cca;
using namespace cca::rci;
using namespace boost;
using namespace rsc;
using namespace rsc::logging;
using namespace rsb;
using namespace rsb::converter;
using namespace rst::converter::rci;

class TestNode: public CCANode {
public:
    TestNode(std::string name) :
        CCANode(name), value() {

        this->addInputPort<cca::rci::JointAngles> ();
        this->addOutputPort<cca::rci::JointAngles> ();

        value = cca::rci::JointAngles(0.0);
    }

    std::string print() const {
        return this->getName();
    }

    void onProcess() {
        if (!this->inputs[0]->empty() && this->inputs[0]->newItem()) {

            cca::rci::JointAnglesPtr step = boost::static_pointer_cast<
                    cca::rci::JointAngles>(this->inputs[0]->tryGet());

            value = cca::rci::JointAngles(
                    value.asDouble(0) + step->asDouble(0));
        }

        cca::rci::JointAngles *tmp = new cca::rci::JointAngles(
                value.asDouble(0));

        // Publish status
        Informer<cca::rci::JointAngles>::DataPtr status(tmp);
        this->outputPorts[0]->publish(status);

        std::cout << "[" << this->getName() << "] Internal status is now:\t"
                << value;
    }

private:
    cca::rci::JointAngles value;
};

/**
 * Simple Control Loop.
 * Listen to Scope /robot/status/nodes/angles/a1 to see the status changing
 * @return
 */
int main() {
    LoggerFactory::getInstance().reconfigure(Logger::LEVEL_WARN);

    /* Beat */
    unsigned int cycletime = 100;
    std::cout << std::endl << "/**" << std::endl
            << " * Setting up the Beat, beating each " << cycletime << " ms."
            << std::endl << " */" << std::endl;
    GlobalBeatPtr heartbeat = GlobalBeat::create(cycletime);
    std::cout << *heartbeat;
    sleep(1);

    /* Nodes */
    std::cout << std::endl << "/**" << std::endl
            << " * Setting up the CCA nodes." << std::endl << " */"
            << std::endl;

    CCANodePtr a1 = CCANodePtr(new TestNode("A1"));
    a1->setProcessingStrategy(TimedProcessing::create(1));
    a1->configureOutputPort(0,
            PortConfiguration::LOCAL("/robot/status/nodes/angles/a1"));
    a1->configureInputPort(0,
            PortConfiguration::LOCAL("/robot/task/nodes/angle"));
    std::cout << *a1;

    CCANodePtr c = CCANodePtr(new Controller<cca::rci::JointAngles> ("Ctrl"));
    c->setProcessingStrategy(TimedProcessing::create(1));
    c->configureInputPort(0, PortConfiguration::LOCAL("/robot/task/angles"));
    c->configureInputPort(1,
            PortConfiguration::LOCAL("/robot/status/nodes/angles/a1"));
    c->configureOutputPort(0,
            PortConfiguration::LOCAL("/robot/task/nodes/angle"));
    std::cout << *c;

    CCANodePtr t = CCANodePtr(
            new TaskProducer<cca::rci::JointAngles> ("uint64_t Task"));
    std::cout << *t;
    t->setProcessingStrategy(TimedProcessing::create(5));
    t->configureOutputPort(0, PortConfiguration::LOCAL("/robot/task/angles"));
    std::cout << *t;

    sleep(1);

    /* Connections */
    std::cout << std::endl << "/**" << std::endl
            << " * Connecting nodes to the heart beat." << std::endl << " */"
            << std::endl;

    heartbeat->registerReceiver(t);
    heartbeat->registerReceiver(c);
    heartbeat->registerReceiver(a1);

    std::cout << std::endl;

    sleep(1);

    /* Beat */
    std::cout << "Runnign control loop ... " << std::endl;
    heartbeat->run();

    return 0;
}
