#include <boost/bind.hpp>
#include <boost/thread.hpp>

#include <rsb/Factory.h>

#include <cca/Beat.h>
#include <cca/processing/RestlessProcessing.h>
#include <cca/processing/TimedProcessing.h>
#include <cca/processing/PortTriggeredProcessing.h>

#include "rci/nodes/ResourceNode.h"
#include "rci/nodes/Controller.h"
#include "rci/nodes/debug/Task.h"
#include "rci/nodes/debug/Status.h"

#include "rci/dto/JointAngles.h"
#include "rci/dto/converter/JointAnglesConverter.h"

using namespace cca;
using namespace cca::rci;
using namespace boost;
using namespace rsc;
using namespace rsc::logging;
using namespace rsb;
using namespace rsb::converter;
using namespace rst::converter::rci;

class TestNode: public CCANode {
public:
    TestNode(std::string name) :
        CCANode(name), value() {

        this->addInputPort<cca::rci::JointAngles> ();
        this->addOutputPort<cca::rci::JointAngles> ();

        value = cca::rci::JointAngles(0.0);
    }

    std::string print() const {
        return this->getName();
    }

    void onProcess() {
        if (!this->inputs[0]->empty() && this->inputs[0]->newItem()) {

            cca::rci::JointAnglesPtr step = boost::static_pointer_cast<
                    cca::rci::JointAngles>(this->inputs[0]->tryGet());

            value = cca::rci::JointAngles(
                    value.asDouble(0) + step->asDouble(0));
        }

        cca::rci::JointAngles *tmp = new cca::rci::JointAngles(
                value.asDouble(0));

        // Publish status
        Informer<cca::rci::JointAngles>::DataPtr status(tmp);
        this->outputPorts[0]->publish(status);

        std::cout << "[" << this->getName() << "] Internal status is now:\t"
                << value;
    }

private:
    cca::rci::JointAngles value;
};

/**
 * Restless on 1
 */
int main() {
    /* Beat */
    unsigned int cycletime = 500;
    std::cout << std::endl << "/**" << std::endl
            << " * Setting up the Beat, beating each " << cycletime << " ms."
            << std::endl << " */" << std::endl;
    GlobalBeatPtr heartbeat;
    heartbeat = GlobalBeat::create(cycletime); // Beat with 50 ms
    std::cout << *heartbeat;
    sleep(1);

    /* Nodes */
    std::cout << std::endl << "/**" << std::endl
            << " * Setting up the CCA nodes." << std::endl << " */"
            << std::endl;
    CCANodePtr a1 = CCANodePtr(new TestNode("A1"));
    a1->setProcessingStrategy(RestlessProcessing::create());
    std::cout << *a1;

    CCANodePtr c = CCANodePtr(new Controller<cca::rci::JointAngles> ("Ctrl"));
    c->setProcessingStrategy(PortTriggeredProcessing::create(1));
    std::cout << *c;

    CCANodePtr t = CCANodePtr(
            new TaskProducer<cca::rci::JointAngles> ("uint64_t Task"));
    t->setProcessingStrategy(TimedProcessing::create(3));
    std::cout << *t;

    CCANodePtr st = CCANodePtr(
            new StatusViewer<cca::rci::JointAngles> ("uint64_t Status"));
    st->setProcessingStrategy(TimedProcessing::create(2));
    std::cout << *st;

    sleep(1);

    /* Connections */
    std::cout << std::endl << "/**" << std::endl
            << " * Connecting nodes to the heart beat." << std::endl << " */"
            << std::endl;

    heartbeat->registerReceiver(t);

    heartbeat->registerReceiver(c);

    heartbeat->registerReceiver(a1);
    // heartbeat->registerReceiver(a2);
    //  heartbeat->registerReceiver(e1);
    //   heartbeat->registerReceiver(a3);
    //   heartbeat->registerReceiver(a4);
    //   heartbeat->registerReceiver(a5);
    //   heartbeat->registerReceiver(a6);
    heartbeat->registerReceiver(st);

    t->configureOutputPort(0, PortConfiguration::LOCAL("/robot/task/angles"));

    a1->configureOutputPort(0,
            PortConfiguration::LOCAL("/robot/status/nodes/angles/a1"));
    /* a2->configureOutputPort(0,"/robot/status/nodes/angles/a2");
     e1->configureOutputPort(0,"/robot/status/nodes/angles/e1");
     a3->configureOutputPort(0,"/robot/status/nodes/angles/a3");
     a4->configureOutputPort(0,"/robot/status/nodes/angles/a4");
     a5->configureOutputPort(0,"/robot/status/nodes/angles/a5");
     a6->configureOutputPort(0,"/robot/status/nodes/angles/a6"); */

    c->configureInputPort(0, PortConfiguration::LOCAL("/robot/task/angles")); // t: /robot/task/angles
    c->configureInputPort(1,
            PortConfiguration::LOCAL("/robot/status/nodes/angles/a1")); // a1: /robot/status/nodes/angles/a1
    c->configureOutputPort(0,
            PortConfiguration::LOCAL("/robot/task/nodes/angle"));

    a1->configureInputPort(0,
            PortConfiguration::LOCAL("/robot/task/nodes/angle")); // c: /robot/task/nodes/angle
    /* a2->configureInputPort(0,"/robot/task/nodes/angle");
     e1->configureInputPort(0,"/robot/task/nodes/angle");
     a3->configureInputPort(0,"/robot/task/nodes/angle");
     a4->configureInputPort(0,"/robot/task/nodes/angle");
     a5->configureInputPort(0,"/robot/task/nodes/angle");
     a6->configureInputPort(0,"/robot/task/nodes/angle");*/

    st->configureInputPort(0,
            PortConfiguration::LOCAL("/robot/status/nodes/angles/a1")); // a1: /robot/status/nodes/angles/a1

    std::cout << std::endl;

    sleep(1);

    /* Beat */
    heartbeat->run();

    return 0;
}
