#pragma once

#include <iostream>

#include "rci/ControlMode.h"
#include "rci/dto/JointAngles.h"
#include "rci/dto/JointTorques.h"
#include "rci/dto/JointImpedance.h"
#include "rci/dto/JointVelocities.h"
#include "rci/dto/Pressure.h"
#include "rci/dto/Length.h"
#include "rci/dto/CartesianPose.h"

namespace rci {

class Controlled {

public:
    /**
     * Controller converged
     *
     * Answers the question, if controller converged after last command.
     */
    virtual bool isConverged() const = 0;

    /**
     * Control mode getter
     *
     * @return Current control mode
     */
    virtual ControlModePtr getControlMode() {
        return this->_controlMode;
    }

    /**
     * Control mode setter
     *
     * @param controlMode Desired control mode
     */
    virtual void setControlMode(ControlModePtr controlMode) {
        this->_controlMode = controlMode;
    }

protected:
    ControlModePtr _controlMode;

    Controlled() :
        _controlMode() {
    }
    virtual ~Controlled() {
    }
};

/**
 * Interface indicating Joint Position Control
 *
 * Interface indicating Joint Position Control on a Node.
 */
class PositionControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setJointPosition(JointAnglesPtr position) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual JointAnglesPtr getLastPositionCommand() const {
        return this->_lastCommandedPosition;
    }

protected:
    PositionControlled() :
        _lastCommandedPosition() {
    }
    virtual ~PositionControlled() {
    }
    JointAnglesPtr _lastCommandedPosition;

};

/**
 * Interface indicating Joint Velocity Control
 *
 * Interface indicating Joint Velocity Control on a Node.
 */
class VelocityControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setJointVelocity(JointVelocitiesPtr velocity) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual JointVelocitiesPtr getLastVelocityCommand() const {
        return this->_lastCommandedVelocity;
    }

protected:
    VelocityControlled() :
        _lastCommandedVelocity() {
    }
    virtual ~VelocityControlled() {
    }
    JointVelocitiesPtr _lastCommandedVelocity;

};

/**
 * Interface indicating Joint Torque Control
 *
 * Interface indicating Joint Torque Control on a Node.
 */
class TorqueControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setJointTorque(JointTorquesPtr position) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual JointTorquesPtr getLastTorqueCommand() const {
        return this->_lastCommandedTorque;
    }

protected:
    TorqueControlled() :
        _lastCommandedTorque() {
    }
    virtual ~TorqueControlled() {
    }
    JointTorquesPtr _lastCommandedTorque;

};

/**
 * Interface indicating Joint Impedance Control
 *
 * Interface indicating Joint Impedance Control on a Node.
 */
class ImpedanceControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setJointImpedance(JointImpedancePtr impedance) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual JointImpedancePtr getLastImpedanceCommand() const {
        return this->_lastCommandedImpedance;
    }

protected:
    ImpedanceControlled() :
        _lastCommandedImpedance() {
    }
    virtual ~ImpedanceControlled() {
    }
    JointImpedancePtr _lastCommandedImpedance;

};

/**
 * Interface indicating Pressure Control on an actuated node
 *
 * Interface indicating Pressure Control on an actuated node.
 */
class PressureControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setPressure(PressurePtr pressure) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual PressurePtr getLastPressureCommand() const {
        return this->_lastCommandedPressure;
    }

protected:
    PressureControlled() :
        _lastCommandedPressure() {
    }
    virtual ~PressureControlled() {
    }
    PressurePtr _lastCommandedPressure;
};

/**
 * Interface indicating Pressure Control on an actuated node
 *
 * Interface indicating Pressure Control on an actuated node.
 */
class LengthControlled {

public:
    /**
     * Sets position
     *
     * Sets a position on a joint/actuator (joint space control).
     */
    virtual bool setLength(LengthPtr Length) = 0;

    /**
     * Returns latest position command. Note, that this is the latest valid
     * commanded position. If the position command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual LengthPtr getLastLengthCommand() const {
        return this->_lastCommandedLength;
    }

protected:
    LengthControlled() :
        _lastCommandedLength() {
    }
    virtual ~LengthControlled() {
    }
    LengthPtr _lastCommandedLength;
};

/**
 * Interface indicating Cartesian Pose Control
 *
 * Interface indicating Cartesian Pose Control on a Node.
 */
class PoseControlled {

public:
    /**
     * Sets a cartesian pose
     *
     * Sets a cartesian pose on a end-effector (task space control).
     */
    virtual bool setPose(PosePtr pose) = 0;

    /**
     * Returns latest pose command. Note, that this is the latest valid
     * commanded pose. If the poe command exceeds limits and is
     * therefore rejected, it ..
     */
    virtual PosePtr getLastPoseCommand() const {
        return this->_lastCommandedPose;
    }

protected:
    PoseControlled() :
        _lastCommandedPose() {
    }

    virtual ~PoseControlled() {
    }

    PosePtr _lastCommandedPose;
};

}
