#pragma once

#include <cca/dto/DataTransferObject.h>

#include "SensingCartesian.h"

#include "rci/dto/JointAngles.h"
#include "rci/dto/CartesianForces.h"
#include "rci/dto/Pressure.h"
#include "rci/dto/Length.h"
#include "rci/dto/JointTorques.h"
#include "rci/dto/Power.h"

namespace rci {

/**
 * Base sensing interface
 */
class Sensing {

public:
	/**
	 * Returns dimensionality of the sensor values.
	 *
	 * @return Dimensionality of the sensor values.
	 */
	virtual unsigned int getDimension() const {
		return this->dimension;
	}

	/**
	 * Update sensor data
	 *
	 * This method might not be required depending on the actual sensor.
	 */
	virtual void update() {
		//
	}

	/**
	 * Read sensor data
	 *
	 * Returns a current readout of the sensor values. This is not wise to use,
	 * use methods with semantics instead.
	 */
	virtual cca::DataTransferObjectPtr get() const {
		throw std::runtime_error("Not implemented.");
	}

protected:
	Sensing(unsigned int dim = 1) :
			dimension(dim) {
	}
	virtual ~Sensing() {
	}
	unsigned int dimension;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense joint angles (encoder value).
 */
class PositionSensing {

public:
	/**
	 * Returns joint position (encoder value)
	 */
	virtual JointAnglesPtr getJointPosition() const = 0;

	/**
	 * Updates internal representation of joint position / angle
	 */
	virtual void updateJointPosition(JointAnglesPtr angle) {
		this->_latestJointPosition = angle;
	}

protected:
	PositionSensing() :
			_latestJointPosition() {
	}
	virtual ~PositionSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	JointAnglesPtr _latestJointPosition;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense joint velocities (encoder value or derived).
 */
class VelocitySensing {

public:
	/**
	 * Returns joint velocity
	 */
	virtual JointAnglesPtr getVelocity() const = 0;

	/**
	 * Updates internal representation of joint position / angle
	 */
	virtual void updateJointVelocity(JointAnglesPtr vel) {
		this->_latestJointVelocity = vel;
	}

protected:
	VelocitySensing() :
			_latestJointVelocity() {
	}
	virtual ~VelocitySensing() {
	}

	/**
	 * Latest sensor read-out
	 * FIXME: Use velocity
	 */
	JointAnglesPtr _latestJointVelocity;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense pressure (for example a vault).
 */
class PressureSensing {

public:
	/**
	 * Returns pressure values
	 */
	virtual PressurePtr getPressure() const {
		return _latestPressureValues;
	}

protected:
	PressureSensing() :
			_latestPressureValues() {
	}
	virtual ~PressureSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	PressurePtr _latestPressureValues;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense a length value.
 */
class LengthSensing {

public:
	/**
	 * Returns length values
	 */
	virtual LengthPtr getLength() const {
		return _latestLengthValues;
	}

protected:
	LengthSensing() :
			_latestLengthValues() {
	}
	virtual ~LengthSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	LengthPtr _latestLengthValues;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense (cartesian) forces.
 */
class ForceSensing {

public:
	/**
	 * Returns absolut forces
	 *
	 * @return 3-dimensional force
	 */
	virtual ForcesPtr getForces() const = 0;

protected:
	ForceSensing() :
			forces() {
	}
	virtual ~ForceSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	ForcesPtr forces;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense joint torques.
 */
class TorqueSensing {

public:
	/**
	 * Returns joint torque
	 *
	 * @return Joint torque
	 */
	virtual JointTorquesPtr getTorque() const {
		if (!torques) {
			throw std::runtime_error(
					"[TorqueSensing::getTorque] No torques received yet!");
		}
		return torques;
	}

	/**
	 * Updates internal representation of torques
	 */
	virtual void updateTorque(JointTorquesPtr torque) {
		if (!torque) {
			throw std::runtime_error(
					"[TorqueSensing::updateTorque] Null Pointer Exception!");
		}
		this->torques->setValue(0, torque->asDouble(0));
	}

protected:
	TorqueSensing() :
			torques() {
	}
	virtual ~TorqueSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	JointTorquesPtr torques;
};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense (cartesian) forces.
 */
class PowerSensing {

public:
	/**
	 * Returns current power consumption
	 *
	 * @return Current power consumption
	 */
	virtual PowerPtr getPower() const = 0;

protected:
	PowerSensing() :
			power() {
	}
	virtual ~PowerSensing() {
	}

	/**
	 * Latest sensor read-out
	 */
	PowerPtr power;
};

}
