#pragma once

#include <cca/dto/DataTransferObject.h>

#include "rci/dto/CartesianPose.h"
#include "rci/dto/CartesianVelocity.h"
#include "rci/dto/CartesianAcceleration.h"

namespace rci {

/**
 * By implementing this interface, a node is characterized as being able to
 * sense a full pose in cartesian space or it`s translational or rotational
 * part.
 */
class OrientationSensing {

public:
    /**
     * Returns absolute Orientation in cartesian task space.
     *
     * @return 3-dimensional task space Orientation
     */
    virtual OrientationPtr getOrientation() const {
        if (!orientation) {
            throw std::runtime_error("No orientation given yet");
        }
        return this->orientation;
    }

    /**
     * Updates internal representation of joint position / angle
     */
    virtual void updateOrientation(OrientationPtr orient) {
        this->orientation = orient;
    }

protected:
    OrientationSensing() :
            orientation() {
    }
    virtual ~OrientationSensing() {
    }

    /**
     * Latest sensor read-out
     */
    OrientationPtr orientation;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense a full pose in cartesian space or it`s translational or rotational
 * part.
 */
class TranslationSensing {

public:

    /**
     * Returns absolute Translation in cartesian task space.
     *
     * @return 3-dimensional task space position
     */
    virtual TranslationPtr getTranslation() const {
        if (!translation) {
            throw std::runtime_error("No translation given yet");
        }
        return this->translation;
    }

    /**
     * Updates internal representation of joint position / angle
     */
    virtual void updateTranslation(TranslationPtr translat) {
        this->translation = translat;
    }

protected:
    TranslationSensing() :
            translation() {
    }
    virtual ~TranslationSensing() {
    }

    /**
     * Latest sensor read-out
     */
    TranslationPtr translation;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense a full pose in cartesian space or it`s translational or rotational
 * part.
 */
class PoseSensing: public TranslationSensing, public OrientationSensing {

public:
    /**
     * Returns absolut Pose in cartesian task space.
     *
     * @return 6-dimensional task space Pose
     */
    virtual PosePtr getPose() const {
        if (!translation) {
            throw std::runtime_error(
                    "Pose incomplete; no translation given yet.");
        }
        if (!orientation) {
            throw std::runtime_error(
                    "Pose incomplete; no orientation given yet.");
        }
        return PosePtr(new Pose(translation, orientation));
    }

    /**
     * Updates internal representation of joint position / angle
     */
    virtual void updatePose(PosePtr pose) {
    	if (pose == NULL) {
    		throw std::runtime_error("[PoseSensing::updatePose] Got null pointer!");
    	}
        this->updateTranslation(pose->getTranslation());
        this->updateOrientation(pose->getOrientation());
    }

    /**
     * Updates internal representation of joint position / angle
     */
    virtual void updatePose(TranslationPtr transl,
            OrientationPtr orient) {
        this->updateTranslation(transl);
        this->updateOrientation(orient);
    }

protected:
    PoseSensing() :
            TranslationSensing(), OrientationSensing() {
    }
    virtual ~PoseSensing() {
    }

};

/**
 * Velocity
 */

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional velocity of an object in cartesian space or it`s
 * translational or rotational part.
 */
class TranslationalVelocitySensing {

public:

    /**
     * Returns translational velocity in cartesian task space.
     *
     * @return 3-dimensional task space position
     */
    virtual TranslationalVelocityPtr getTranslationalVelocity() const {
        if (!veltranslational) {
            throw std::runtime_error("No translational velocity given yet");
        }
        return this->veltranslational;
    }

    /**
     * Updates sensor value for the translational part of the measured cartesian
     * velocity.
     *
     * @param tvel
     *          Translational velocity
     */
    virtual void updateTranslationalVelocity(TranslationalVelocityPtr tvel) {
        this->veltranslational = tvel;
    }

protected:
    TranslationalVelocitySensing() :
            veltranslational() {
    }
    virtual ~TranslationalVelocitySensing() {
    }

    /**
     * Latest sensor read-out
     */
    TranslationalVelocityPtr veltranslational;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional velocity of an object in cartesian space or it`s
 * translational or rotational part.
 *
 * @todo Also update and return CartesianVelocity (joined transl. and rotat.)
 */
class RotationalVelocitySensing {

public:

    /**
     * Returns rotational velocity in cartesian task space.
     *
     * @return 3-dimensional task space Orientation
     */
    virtual RotationalVelocityPtr getRotationalVelocity() const {
        if (!velrotational) {
            throw std::runtime_error("No rotational velocity given yet");
        }
        return this->velrotational;
    }

    /**
     * Updates sensor value for the rotational part of the measured cartesian
     * velocity.
     *
     * @param rvel
     *          Rotational velocity
     */
    virtual void updateRotationalVelocity(RotationalVelocityPtr rvel) {
        this->velrotational = rvel;
    }

protected:
    RotationalVelocitySensing() :
            velrotational() {
    }
    virtual ~RotationalVelocitySensing() {
    }

    /**
     * Latest sensor read-out
     */
    RotationalVelocityPtr velrotational;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional velocity of an object in cartesian space or it`s
 * translational or rotational part.
 *
 * @todo Also update and return CartesianVelocity (joined transl. and rotat.)
 */
class CartesianVelocitySensing: public TranslationalVelocitySensing,
        public RotationalVelocitySensing {

public:

    /**
     * Updates sensor value for the measured cartesian velocity.
     *
     * @param transvel
     *          Translational velocity
     * @param rotvel
     *          Rotational velocity
     */
    virtual void updateVelocity(TranslationalVelocityPtr transvel,
            RotationalVelocityPtr rotvel) {
        this->updateTranslationalVelocity(transvel);
        this->updateRotationalVelocity(rotvel);
    }

    virtual VelocityPtr getVelocity() {
        throw std::runtime_error("Not yet implemented");
    }

protected:
    CartesianVelocitySensing() :
            TranslationalVelocitySensing(), RotationalVelocitySensing() {
    }
    virtual ~CartesianVelocitySensing() {
    }

};

/**
 * Acceleration
 */

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional acceleration of an object in cartesian space or it`s
 * translational or rotational part.
 */
class TranslationalAccelerationSensing {

public:

    /**
     * Returns translational acceleration in cartesian task space.
     *
     * @return 3-dimensional task space position
     */
    virtual TranslationalAccelerationPtr getTranslationalAcceleration() const {
        if (!acctranslational) {
            throw std::runtime_error("No translational acceleration given yet");
        }
        return this->acctranslational;
    }

    /**
     * Updates sensor value for the translational part of the measured cartesian
     * acceleration.
     *
     * @param tacc
     *          Translational acceleration
     */
    virtual void updateTranslationalAcceleration(TranslationalAccelerationPtr tacc) {
        this->acctranslational = tacc;
    }

protected:
    TranslationalAccelerationSensing() :
            acctranslational() {
    }
    virtual ~TranslationalAccelerationSensing() {
    }

    /**
     * Latest sensor read-out
     */
    TranslationalAccelerationPtr acctranslational;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional acceleration of an object in cartesian space or it`s
 * translational or rotational part.
 *
 * @todo Also update and return CartesianAcceleration (joined transl. and rotat.)
 */
class RotationalAccelerationSensing {

public:

    /**
     * Returns rotational acceleration in cartesian task space.
     *
     * @return 3-dimensional task space Orientation
     */
    virtual RotationalAccelerationPtr getRotationalAcceleration() const {
        if (!accrotational) {
            throw std::runtime_error("No rotational acceleration given yet");
        }
        return this->accrotational;
    }

    /**
     * Updates sensor value for the rotational part of the measured cartesian
     * acceleration.
     *
     * @param racc
     *          Rotational acceleration
     */
    virtual void updateRotationalAcceleration(RotationalAccelerationPtr racc) {
        this->accrotational = racc;
    }

protected:
    RotationalAccelerationSensing() :
            accrotational() {
    }
    virtual ~RotationalAccelerationSensing() {
    }

    /**
     * Latest sensor read-out
     */
    RotationalAccelerationPtr accrotational;

};

/**
 * By implementing this interface, a node is characterized as being able to
 * sense the 6-dimensional acceleration of an object in cartesian space or it`s
 * translational or rotational part.
 *
 * @todo Also update and return CartesianAcceleration (joined transl. and rotat.)
 */
class CartesianAccelerationSensing: public TranslationalAccelerationSensing,
        public RotationalAccelerationSensing {

public:

    /**
     * Updates sensor value for the measured cartesian acceleration.
     *
     * @param transacc
     *          Translational acceleration
     * @param rotacc
     *          Rotational acceleration
     */
    virtual void updateAcceleration(TranslationalAccelerationPtr transacc,
            RotationalAccelerationPtr rotacc) {
        this->updateTranslationalAcceleration(transacc);
        this->updateRotationalAcceleration(rotacc);
    }

    virtual AccelerationPtr getAcceleration() {
        throw std::runtime_error("Not yet implemented");
    }

protected:
    CartesianAccelerationSensing() :
            TranslationalAccelerationSensing(), RotationalAccelerationSensing() {
    }
    virtual ~CartesianAccelerationSensing() {
    }

};

}
