#pragma once

#include <vector>
#include <iostream>
#include <boost/shared_ptr.hpp>

#include "rci/ResourceNode.h"

namespace rci {

class Synchronizer;
typedef boost::shared_ptr<Synchronizer> SynchronizerPtr;

/**
 * Synchronizer
 * Different synchronizers may be for example:
 * * Update with a given frequency
 * ** Send data every time
 * ** Send only if new data
 * * Update only if new data (event-based)
 * * Triggered from robot side (e.g. FRI)
 */
class Synchronizer {
public:
    Synchronizer(const std::string &name);
    virtual ~Synchronizer();

    /**
     * Synchronous processing (synchronisation of commands and sensor values
     * with the robot). After calling this, you can safely access new sensor
     * values.
     * Generic implementation of this method calls processAsync and then
     * waitForProcessAsync.
     */
    virtual void process();

    /**
     * Asynchronous processing (synchronisation of commands and sensor values
     * with the robot). This method will return immediately, you will have to
     * waitForProcessAsync to finish to have new sensor values and safely access
     * them.
     */
    virtual void processAsync() = 0;

    /**
     * Waits until asynchronous processing (synchronisation of commands and
     * sensor values with the robot) is done.
     * You will need for this to know that new sensor values arrived and you can
     * read them savely.
     */
    virtual void waitForProcessAsync() = 0;

    /**
     * Starts synchronizing. For some backends this has to be stopped, e.g. for
     * control mode switches.
     */
    virtual bool start();

    /**
     * Stops synchronizing, e.g. to have time for control mode switches.
     */
    virtual bool stop();

    /**
     * Returns, if the synchronizer is (too) busy right now. This
     * is important to know in case of robots where control has to be paused in
     * order to make a certain action (e.g. control mode switch).
     */
    virtual bool tooBusy() const;

    /**
     *
     */
    friend std::ostream
    & operator<<(std::ostream& os, const Synchronizer& val);

protected:

    /**
     * Only true for the first time
     */
    bool _onlyReceive;

    /**
     *
     */
    bool busy;

    /**
     *
     */
    std::string name;
};

}
