#include "CartesianPose.h"

#include <math.h>

#include <boost/math/quaternion.hpp>

using namespace std;
using namespace nemo;
using namespace rsc::math;

namespace rci {

Translation::Translation() :
        CartesianTranslational() {
}

Translation::Translation(double valx, double valy, double valz) :
        CartesianTranslational(valx, valy, valz) {
}

Translation::Translation(const Translation& transl) :
        CartesianTranslational(transl) {
}

Translation::Translation(const TranslationPtr& transl) :
        CartesianTranslational(transl) {
}

TranslationPtr Translation::fromMeters(double x, double y, double z) {
    return TranslationPtr(new Translation(x, y, z));
}

Translation::Translation(const RealVector& values) :
        CartesianTranslational(values) {
}

std::string Translation::print() const {
    ostringstream outstream(ostringstream::out);
    outstream.precision(3); // Precision when printing double values
    outstream << "Translation<x=" << this->asDouble(0) << ",y="
            << this->asDouble(1) << ",z=" << this->asDouble(2) << ">(m)"
            << std::endl;
    return outstream.str();
}

Orientation::Orientation() :
        CartesianRotational() {
}

Orientation::Orientation(double vala, double valb, double valc, double vald) :
        CartesianRotational(vala, valb, valc, vald) {
}

Orientation::Orientation(const RealVector& values) :
        CartesianRotational(values) {
}

Orientation::Orientation(const Orientation& orient) :
        CartesianRotational(orient) {
}

Orientation::Orientation(const OrientationPtr& orient) :
        CartesianRotational(orient) {
}

Orientation::Orientation(const ::boost::math::quaternion<double>& quaternion) :
        CartesianRotational(quaternion) {
}

Orientation::Orientation(double rxx, double rxy, double rxz, double ryx,
        double ryy, double ryz, double rzx, double rzy, double rzz) :
        CartesianRotational(rxx, rxy, rxz, ryx, ryy, ryz, rzx, rzy, rzz) {
}

OrientationPtr Orientation::rotateBy(const OrientationPtr& rot) const {
    return Orientation::fromQuaternion(this->_quaternion * rot->asQuaternion());
}

std::string Orientation::print() const {
    ostringstream outstream(ostringstream::out);
    outstream.precision(3); // Precision when printing double values
    outstream << "Orientation<q0=" << this->q0() << ",q1=" << this->q1()
            << ",q2=" << this->q2() << ",q3=" << this->q3() << ">" << std::endl;
    return outstream.str();
}

OrientationPtr Orientation::fromEulerAngles(double rotX, double rotY,
        double rotZ) {
    double q0 = cos(rotX / 2) * cos(rotY / 2) * cos(rotZ / 2)
            + sin(rotX / 2) * sin(rotY / 2) * sin(rotZ / 2);
    double q1 = sin(rotX / 2) * cos(rotY / 2) * cos(rotZ / 2)
            - cos(rotX / 2) * sin(rotY / 2) * sin(rotZ / 2);
    double q2 = cos(rotX / 2) * sin(rotY / 2) * cos(rotZ / 2)
            + sin(rotX / 2) * cos(rotY / 2) * sin(rotZ / 2);
    double q3 = cos(rotX / 2) * cos(rotY / 2) * sin(rotZ / 2)
            - sin(rotX / 2) * sin(rotY / 2) * cos(rotZ / 2);

    return OrientationPtr(new Orientation(q0, q1, q2, q3));
}

OrientationPtr Orientation::fromAxisAngle(double angle, double vecX,
        double vecY, double vecZ) {

    if (fabs(angle) < epsilon()) {
        // Angle is zero, direction vector has no effect
        return OrientationPtr(new Orientation(1.0, 0.0, 0.0, 0.0));
    }

    double qx = vecX * sin(angle / 2);
    double qy = vecY * sin(angle / 2);
    double qz = vecZ * sin(angle / 2);
    double qw = cos(angle / 2);

    return OrientationPtr(new Orientation(qw, qx, qy, qz));
}

OrientationPtr Orientation::fromCompactAxisAngle(double aaX, double aaY,
        double aaZ) {
    double angle = sqrt(aaX * aaX + aaY * aaY + aaZ * aaZ);
    return Orientation::fromAxisAngle(angle, aaX / angle, aaY / angle,
            aaZ / angle);
}

OrientationPtr Orientation::fromQuaternion(double q0, double q1, double q2,
        double q3) {
    ::boost::math::quaternion<double> quaternion(q0, q1, q2, q3);
    return OrientationPtr(new Orientation(quaternion));
}

OrientationPtr Orientation::fromQuaternion(
        const ::boost::math::quaternion<double>& quaternion) {
    return OrientationPtr(new Orientation(quaternion));
}

OrientationPtr Orientation::fromRotationMatrix(double rxx, double rxy,
        double rxz, double ryx, double ryy, double ryz, double rzx, double rzy,
        double rzz) {
    return OrientationPtr(
            new Orientation(rxx, rxy, rxz, ryx, ryy, ryz, rzx, rzy, rzz));
}

Pose::Pose() :
        CartesianValue() {
}

Pose::Pose(double valx, double valy, double valz, double vala, double valb,
        double valc, double vald) :
        CartesianValue(valx, valy, valz, vala, valb, valc, vald) {
}

Pose::Pose(double valx, double valy, double valz, double vala, double valb,
        double valc) :
        CartesianValue(
                CartesianTranslationalPtr(
                        new CartesianTranslational(valx, valy, valz)),
                CartesianRotationalPtr(
                        new CartesianRotational(vala, valb, valc))) {
}

Pose::Pose(const Pose& pose) :
        CartesianValue(pose) {
}

Pose::Pose(const RealVector& values) :
        CartesianValue(values) {
}

Pose::Pose(const TranslationPtr& transl, const OrientationPtr& rot) :
        CartesianValue(transl, rot) {
}

TranslationPtr Pose::getTranslation() const {
    return Translation::fromMeters(this->x(), this->y(), this->z());
}

OrientationPtr Pose::getOrientation() const {
    return Orientation::fromQuaternion(this->q0(), this->q1(), this->q2(),
            this->q3());
}

double Pose::x() const {
    return this->_transl.x();
}

double Pose::y() const {
    return this->_transl.y();
}

double Pose::z() const {
    return this->_transl.z();
}

double Pose::a() const {
    return this->_rot.a();
}

double Pose::b() const {
    return this->_rot.b();
}

double Pose::c() const {
    return this->_rot.c();
}

double Pose::q0() const {
    return this->_rot.q0();
}

double Pose::q1() const {
    return this->_rot.q1();
}

double Pose::q2() const {
    return this->_rot.q2();
}

double Pose::q3() const {
    return this->_rot.q3();
}

PosePtr Pose::fromMetersAndRadians(double x, double y, double z, double q0,
        double q1, double q2, double q3) {
    PosePtr pose(new Pose(x, y, z, q0, q1, q2, q3));
    return pose;
}

PosePtr Pose::fromMetersAndRadians(double x, double y, double z, double a,
        double b, double c) {
    TranslationPtr transl = Translation::fromMeters(x, y, z);
    OrientationPtr rot = boost::static_pointer_cast<Orientation>(
            Orientation::fromEulerAngles(a, b, c));
    return PosePtr(new Pose(transl, rot));
}

PosePtr Pose::fromMetersAndRadiansHomogeneous(double rxx, double rxy,
        double rxz, double tx, double ryx, double ryy, double ryz, double ty,
        double rzx, double rzy, double rzz, double tz) {
    TranslationPtr transl = Translation::fromMeters(tx, ty, tz);
    OrientationPtr rot = boost::static_pointer_cast<Orientation>(
            Orientation::fromRotationMatrix(rxx, rxy, rxz, ryx, ryy, ryz, rzx,
                    rzy, rzz));
    return PosePtr(new Pose(transl, rot));
}

PosePtr Pose::fromMetersAndRadiansHomogeneous(const RealMatrix& homogeneous) {
    return Pose::fromMetersAndRadiansHomogeneous(homogeneous(0, 0),
            homogeneous(0, 1), homogeneous(0, 2), homogeneous(0, 3),
            homogeneous(1, 0), homogeneous(1, 1), homogeneous(1, 2),
            homogeneous(1, 3), homogeneous(2, 0), homogeneous(2, 1),
            homogeneous(2, 2), homogeneous(2, 3));
}

std::string Pose::print() const {
    ostringstream outstream(ostringstream::out);
    outstream.precision(3); // Precision when printing double values
    outstream << "Pose<" << this->_transl << " , " << this->_rot << ">";
    return outstream.str();
}

}
