#pragma once

#include <iostream>

#include <boost/shared_ptr.hpp>
#include <boost/math/quaternion.hpp>

#include "cca/dto/DataTransferObject.h"
#include "cca/dto/ContainingDoubles.h"

namespace rci {

class CartesianTranslational;
typedef boost::shared_ptr<CartesianTranslational> CartesianTranslationalPtr;
class CartesianRotational;
typedef boost::shared_ptr<CartesianRotational> CartesianRotationalPtr;
class CartesianValue;
typedef boost::shared_ptr<CartesianValue> CartesianValuePtr;

/**
 * Domain object, representing a position or translational transformation in
 * cartesian space.
 * @ingroup DTO
 */
class CartesianTranslational: public cca::DataTransferObject,
        public cca::ContainingDoubles {

public:
    CartesianTranslational();
    CartesianTranslational(double valx, double valy, double valz);
    CartesianTranslational(const nemo::RealVector& values);
    CartesianTranslational(const CartesianTranslational& transl);
    CartesianTranslational(const CartesianTranslationalPtr& transl);
    virtual ~CartesianTranslational();

    /**
     * Domain specific getter, return x position
     *
     * @return x position
     */
    virtual double x() const;

    /**
     * Domain specific getter, return y position
     * @return y position
     */
    virtual double y() const;

    /**
     * Domain specific getter, return z position
     * @return z position
     */
    virtual double z() const;

    /**
     * Printing the translation object
     *
     * @return Textual representation of this translation.
     */
    virtual std::string print() const;
};

/**
 * Domain object, representing an orientation or rotational transformation in
 * cartesian space.
 *
 * @ingroup DTO
 */
class CartesianRotational: public cca::DataTransferObject,
        public cca::ContainingDoubles {

public:
    CartesianRotational();
    CartesianRotational(double vala, double valb, double valc);
    CartesianRotational(double vala, double valb, double valc, double vald);
    CartesianRotational(const nemo::RealVector& values);
    CartesianRotational(const CartesianRotational& rot);
    CartesianRotational(const CartesianRotationalPtr& rot);
    CartesianRotational(const ::boost::math::quaternion<double> quaternion);
    CartesianRotational(double rxx, double rxy, double rxz, double ryx,
            double ryy, double ryz, double rzx, double rzy, double rzz);
    ~CartesianRotational();

    /**
     * Domain specific getter, return a rotation
     * @return a rotation
     */
    virtual double a() const;

    /**
     * Domain specific getter, return b rotation
     * @return b rotation
     */
    virtual double b() const;

    /**
     * Domain specific getter, return c rotation
     * @return c rotation
     */
    virtual double c() const;

    /**
     * Domain specific getter, return a rotation
     * @return a rotation
     */
    virtual double q0() const;

    /**
     * Domain specific getter, return b rotation
     * @return b rotation
     */
    virtual double q1() const;

    /**
     * Domain specific getter, return c rotation
     * @return c rotation
     */
    virtual double q2() const;

    /**
     * Domain specific getter, return c rotation
     * @return c rotation
     */
    virtual double q3() const;

    virtual nemo::RealVector asDoubleVector() const;

    /**
     * Returns rotation components (quaternion components) \f$q0, q1, q2, q3\f$
     *
     * @param index
     * 			Index to receive
     * @return Rotation component (quaternion component)
     */
    virtual double asDouble(unsigned int index) const;

    /// This setter is illegal for cartesian orientation
    virtual void setValue(double value);

    /**
     * Sets values in its default representation (quaternions).
     *
     * @param index
     *          Quaternion index to set the value
     * @param value
     *          Desired value
     */
    virtual void setValue(unsigned int index, double value);

    /**
     * Returns orientation as quaternion.
     * @return Orientation represented as quaternion.
     */
    virtual ::boost::math::quaternion<double> asQuaternion() const;

    /**
     * Returns orientation in euler angles. Convention is euler angles in
     * order x, y, z - right-handed.
     * @return Vector with the three euler angles: x, y, z - right-handed.
     */
    virtual nemo::RealVector asEulerAngles() const;

    /**
     * Returns orientation in axis-angle representation: angle, axis as
     * unit vector {x, y, z}.
     * @return Vector: angle, x, y, z
     */
    virtual nemo::RealVector asAxisAngle() const;

    /**
     * Returns orientation in compact axis-angle representation: Axis {x, y, z},
     * rotation angle is encoded in the norm.
     * @return Vector: x, y, z
     */
    virtual nemo::RealVector asCompactAxisAngle() const;

    /**
     * Printing the cartesian rotational object
     * @return Textual representation of this object.
     */
    virtual std::string print() const;

protected:

    /// Epsilon, threshold for detecting singularities in conversions
    static double epsilon() {
        return 10e-7;
    }

    /**
     * Signum of a double value
     *
     * @param value
     *          Arbitrary double value.
     *
     * @return Signum (+1 if value >0, -1 if value <0, 0 if value==0)
     */
    static double signum(double value) {
        if (value > 0)
            return 1;
        if (value < 0)
            return -1;
        return 0;
    }

    /**
     * Quaternion to store the rotation
     */
    ::boost::math::quaternion<double> _quaternion;
};

/**
 * Domain object, representing a full cartesian pose, consisting of
 * translational transformation and orientation.
 * @ingroup DTO
 */
class CartesianValue: public cca::DataTransferObject {

public:
    CartesianTranslational _transl;
    CartesianRotational _rot;

    CartesianValue();
    CartesianValue(double valx, double valy, double valz, double vala,
            double valb, double valc, double vald);
    CartesianValue(const nemo::RealVector& values);
    CartesianValue(const CartesianValue& val);
    CartesianValue(const CartesianTranslationalPtr& transl,
            const CartesianRotationalPtr& rot);
    virtual ~CartesianValue();

    /**
     * Returns translation and rotation components (quaternion components):
     * \f$x, y, z, q0, q1, q2, q3\f$
     *
     * @param index
     * 			Index to receive
     * @return Translation or rotation component
     */
    virtual double asDouble(unsigned int index) const;

    /**
     * Returns translation and rotation components (quaternion components) as
     * 7-dimensional vector \f$x, y, z, q0, q1, q2, q3\f$
     *
     * @return 7-dimensional vector \f$x, y, z, q0, q1, q2, q3\f$
     */
    virtual nemo::RealVector asDoubleVector() const;

    /**
     * Sets double values (in-place), index 0-2 set Translation in default
     * representation, index 3-6 sets Orientation in default representation.
     *
     * @param index
     *          Index to set value at
     * @param value
     *          Desired value
     */
    virtual void setValue(unsigned int index, double value);

    /**
     * Domain specific getter for 1st translatory value.
     *
     * @return 1st translatory value
     */
    virtual double x() const;

    /**
     * Domain specific getter for 2nd translatory value.
     *
     * @return 2nd translatory value
     */
    virtual double y() const;

    /**
     * Domain specific getter for 3rd translatory value.
     *
     * @return 3rd translatory value
     */
    virtual double z() const;

    /**
     * Domain specific getter for 1st rotatory value.
     *
     * @return 1st rotatory value
     */
    virtual double a() const;

    /**
     * Domain specific getter for 2nd rotatory value.
     *
     * @return 2nd rotatory value
     */
    virtual double b() const;

    /**
     * Domain specific getter for 3rd rotatory value.
     *
     * @return 3rd rotatory value
     */
    virtual double c() const;

    /**
     * Domain specific getter for 1st rotatory value.
     *
     * @return 1st rotatory value
     */
    virtual double q0() const;

    /**
     * Domain specific getter for 2nd rotatory value.
     *
     * @return 2nd rotatory value
     */
    virtual double q1() const;

    /**
     * Domain specific getter for 3rd rotatory value.
     *
     * @return 3rd rotatory value
     */
    virtual double q2() const;

    /**
     * Domain specific getter for 3rd rotatory value.
     *
     * @return 3rd rotatory value
     */
    virtual double q3() const;

    /**
     * Get dimension / number of values
     *
     * @return Number of contained values (6)
     */
    virtual unsigned int getDimension() const;
};

}
