#pragma once

#include <iostream>
#include <sstream>

// RSC
#include <rsc/math/MathUtils.h>

#include "rci/dto/CartesianPose.h"

namespace rci {

class Velocity;
typedef boost::shared_ptr<Velocity> VelocityPtr;
class TranslationalVelocity;
typedef boost::shared_ptr<TranslationalVelocity> TranslationalVelocityPtr;
class RotationalVelocity;
typedef boost::shared_ptr<RotationalVelocity> RotationalVelocityPtr;

/**
 * Domain object, representing a cartesian
 * translatory velocity. Default representation is in \f$\frac{m}{s}\f$,
 * according to the International Unit System (SI).
 * @ingroup DTO
 */
class TranslationalVelocity: public CartesianTranslational {

public:

    TranslationalVelocity();
    TranslationalVelocity(double valx, double valy, double valz);
    TranslationalVelocity(nemo::RealVector values);
    ~TranslationalVelocity();

    virtual double m_s(unsigned int index);
    virtual nemo::RealVector m_sVector();

    /**
     * Shared Pointer Factory
     */
    static TranslationalVelocityPtr fromM_s(double x, double y, double z);
protected:
    /**
     * Printing the pose object
     * @return Textual representation of this pose.
     */
    virtual std::string print() const;
};

/**
 * Domain object, representing an rotational velocity in
 * cartesian space as quaternions (for reference see
 * http://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation ).
 * Default representation is in \f$\frac{m}{s}\f$, according to the
 * International Unit System (SI).
 *
 * @ingroup DTO
 */
class RotationalVelocity: public CartesianRotational {

public:
    RotationalVelocity();
    RotationalVelocity(double q0, double q1, double q2, double q3);
    RotationalVelocity(nemo::RealVector values);
    RotationalVelocity(const RotationalVelocity& orientation);
    RotationalVelocity(const ::boost::math::quaternion<double> quaternion);
    virtual ~RotationalVelocity();

    /**
     * Create orientation from given euler angles in \f$\frac{m}{s}\f$.
     * Convention is euler angles in order x, y, z - right-handed.
     */
    static RotationalVelocityPtr
    fromEulerAngles(double rotX, double rotY, double rotZ);

    /**
     * Create orientation from given values in axis-angle representation in
     * \f$\frac{m}{s}\f$.
     * Expects rotation angle and axis as unit vector {x, y, z}.
     * @param angle Rotation angle
     * @param vecX X-Component of rotation axis
     * @param vecY Y-Component of rotation axis
     * @param vecZ Z-Component of rotation axis
     */
    static RotationalVelocityPtr
    fromAxisAngle(double angle, double vexX, double vecY, double vecZ);

    /**
     * Create orientation from given values in compact axis-angle
     * representation in \f$\frac{m}{s}\f$.
     * Expects rotation as axis {x, y, z}, multiplied by rotation angle.
     */
    static RotationalVelocityPtr
    fromCompactAxisAngle(double x, double y, double z);

    /**
     * Create orientation from
     */
    static RotationalVelocityPtr
    fromQuaternion(double q0, double q1, double q2, double q3);

    /**
     * Create orientation from
     */
    static RotationalVelocityPtr
    fromQuaternion(const ::boost::math::quaternion<double> quaternion);

protected:
    /**
     * Printing the orientation object
     * @return Textual representation of this orientation.
     */
    virtual std::string print() const;
};

}
