#pragma once

#include <iostream>
#include <sstream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include <nemo/Mapping.h>

#include <rsc/math/MathUtils.h>

#include "rci/dto/JointValues.h"

namespace rci {

class JointAngles;
typedef boost::shared_ptr<JointAngles> JointAnglesPtr;

/**
 * Domain object, representing Joint Angles.
 * Default representatin is in radian (\f$rad\f$), according to the International
 * System of Units (SI).
 *
 * @ingroup DTO
 */
class JointAngles: public JointValues {

public:
    JointAngles();
    JointAngles(double value);
    JointAngles(const nemo::RealVector& values);
    JointAngles(const JointAngles&);

    /**
     * Returns single joint value in default representation (\f$rad\f$)
     *
     * @param index
     * 			Index of the joint value
     * @param normalize
     * 			Normalization of angle to interval \f$[0,2\pi[\f$
     * @return Joint angle
     */
    virtual double asDouble(unsigned int index, bool norm = false) const;

    /**
     * Returns single joint value in \f$rad\f$
     *
     * @param index
     * 			Index of the joint value
     * @param normalize
     * 			Normalization of angle to interval \f$[0,2\pi[\f$
     * @return Joint angle (rad)
     */
    virtual double rad(unsigned int index = 0, bool norm = false) const;

    /**
     * Returns vector of joint values in \f$rad\f$
     *
     * @param normalize
     * 			Normalization of angle to interval \f$[0,2\pi[\f$
     * @return Joint angles
     */
    virtual nemo::RealVector radVector(bool norm = false) const;

    /**
     * Returns single joint value in degree
     *
     * @param index
     * 			Index of the joint value
     * @param normalize
     * 			Normalization of angle to interval \f$[0,360[\f$
     * @return Joint angles
     */
    virtual double deg(unsigned int index = 0, bool norm = false) const;

    /**
     * Returns vector of joint values in degree
     *
     * @param normalize
     * 			Normalization of angle to interval \f$[0,360[\f$
     * @return Joint angles
     */
    virtual nemo::RealVector degVector(bool norm = false) const;

    /**
     * Returns single joint value
     *
     * @param index
     * 			Index of the joint value
     * @param normalize
     * 			Normalization of angle to interval \f$[0,2\pi[\f$
     * @return Joint angles as vector
     */
    virtual nemo::RealVector asDoubleVector(bool norm = false) const;

    /**
     * Set the values of this JointAngle object (in \f$rad\f$)
     *
     * @param index
     * @param value
     */
    virtual void setFromRad(unsigned int index, double value);
    virtual void setFromRad(const nemo::RealVector& values);

    /**
     * Set the values of this JointAngle object (in degree)
     *
     * @param index
     * @param value
     */
    virtual void setFromDeg(unsigned int index, double value);
    virtual void setFromDeg(const nemo::RealVector& values);

    /**
     * Create blank JointAnglesPtr with given dimension (and optionally fill it
     * with the given value, otherwise filled with zeros).
     *
     * @param dimension
     *          Dimension of the JointAngles DTo to create
     * @param value
     *          Value to fill each end every field of the DTO (value in default
     *          representation - \f$rad\f$)
     */
    static JointAnglesPtr create(unsigned int dimension, double value = 0.0);

    /**
     * Create JointAngles from a JointAngles object (deep copy)
     *
     * @param JointAngles object
     */
    static JointAnglesPtr copy(const JointAngles&);

    /**
     * Create JointAngles from a vector of radian values.
     *
     * @param values
     * 			Vector of values in radian
     */
    static JointAnglesPtr fromRad(const nemo::RealVector& values);
    static JointAnglesPtr fromRad(double value);

    /**
     * Create JointAngles from a vector of degree values.
     *
     * @param values
     * 			Vector of values in degree
     */
    static JointAnglesPtr fromDeg(const nemo::RealVector& values);
    static JointAnglesPtr fromDeg(double value);

    /**
     * Normalization of all values to interval \f$[0,2\pi[\f$
     */
    void normalize();

    /**
     * Factory
     */
    static JointAnglesPtr fromValue(double value);

    /**
     * Print function
     */
    virtual std::string print() const;
};

}
