#pragma once

#include "rci/dto/JointValues.h"

namespace rci {

class JointImpedance;
typedef boost::shared_ptr<JointImpedance> JointImpedancePtr;
class JointStiffness;
typedef boost::shared_ptr<JointStiffness> JointStiffnessPtr;
class JointDamping;
typedef boost::shared_ptr<JointDamping> JointDampingPtr;

/**
 * Domain object, representing JointStiffness, default representation is in
 * \f$[N*m/rad]\f$.
 *
 * @ingroup DTO
 * @ingroup Compliance
 */
class JointStiffness: public JointValues {

public:
    /**
     * Empty constructor
     * Initializes JointImpedance with no values.
     */
    JointStiffness();

    /**
     * Constructor from double vector.
     * Initializes Joint stiffness with given values from double vector.
     *
     * @param values
     *          Double vector with stiffness values in \f$[N*m/rad]\f$
     */
    JointStiffness(const nemo::RealVector& values);

    /**
     * Constructor from single stiffness value.
     * Initializes Joint stiffness for one single joint.
     *
     * @param stiffness
     *          Double with stiffness value in \f$[N*m/rad]\f$
     */
    JointStiffness(double stiffness);

    /**
     * Constructor from dimension and stiffness value.
     * Initializes JointStiffness of given dimension with given stiffness value.
     *
     * @param dim
     * 			Dimension of constructed object Stiffness
     * @param stiffness
     *          Double with stiffness value in \f$[N*m/rad]\f$
     */
    JointStiffness(unsigned int dim, double stiffness);

    /**
     *
     * @param index
     * @return
     */
    JointStiffness operator[](unsigned int index) const;

protected:
    /**
     * Print function
     */
    virtual std::string print() const;
};

/**
 * Domain object, representing JointDamping, default representation is in
 * ???.
 *
 * @todo what is SI metric unit for rotatory damping?
 *
 * @ingroup DTO
 * @ingroup Compliance
 */
class JointDamping: public JointValues {

public:
    /**
     * Empty constructor
     * Initializes JointImpedance with no values.
     */
    JointDamping();

    /**
     * Constructor from double vector.
     * Initializes Joint damping with given values from double vector.
     *
     * @param values
     *          Double vector with damping values in ????
     */
    JointDamping(const nemo::RealVector& values);

    /**
     * Constructor from single damping value.
     * Initializes Joint damping for one single joint.
     *
     * @param damping
     *          Stiffness value
     */
    JointDamping(double damping);

    /**
     * Constructor from single damping value.
     * Initializes Joint damping for one single joint.
     *
     * @param dim
     *          Dimension of the constrcuted object
     * @param damping
     *          Stiffness value for all dimensions
     */
    JointDamping(unsigned int dim, double damping);

    /**
     * Returns one-dimensional
     *
     * @param index
     *
     * @return
     */
    JointDamping operator[](unsigned int index) const;

protected:
    /**
     * Print function
     */
    virtual std::string print() const;
};

/**
 * Domain object, representing JointImpedance, consisting of JointStiffness and
 * JointDamping.
 *
 * @ingroup DTO
 * @ingroup Compliance
 */
class JointImpedance: public JointValues {

public:
    /**
     * Empty constructor
     * Initializes JointImpedance with no values.
     */
    JointImpedance();

    /**
     * Initializes one-dimensional JointImpedance with given stiffness and
     * damping.
     */
    JointImpedance(double stiffness, double damping);

    /**
     * Constructor from double vector.
     * Initializes JointImpedance with given values from double vector. Vector
     * has to be twice the number of joints, stiffness and damping alternating.
     *
     * @param values
     *          Double vector with stiffness and damping values (in that
     *          order, units are (\f$[N*m/rad]\f$, ..., ???)
     *
     * @todo what is SI metric unit for rotatory damping?
     */
    JointImpedance(const nemo::RealVector& values);

    /**
     * Constructor from double vector.
     * Initializes JointImpedance with given values from double vector. Vector
     * has to be twice the number of joints, stiffness and damping alternating.
     *
     * @param stiffness
     *          Double vector with stiffness in \f$[N*m/rad]\f$
     * @param damping
     *          Double vector with stiffness in  ???
     */
    JointImpedance(const nemo::RealVector& stiffness,
            const nemo::RealVector& damping);

    /**
     * Constructor from JointStiffness and JointDamping. Dimensions if stiffness
     * and damping must agree.
     *
     * @param stiffness
     * 			Joint stiffness
     * @param damping
     *          Joint damping
     */
    JointImpedance(JointStiffness stiffness, JointDamping damping);

    /**
     * Copy from JointImpedance.
     *
     * @param impedance
     *          JointImpedance to copy
     *
     * @return Shared pointer to the copy
     */
    static JointImpedancePtr copy(const JointImpedance& impedance);

    /**
     * Crate JointImpedancePtr from values vector.
     *
     * @param values
     *          Vector of raw values.
     *
     * @return Newly created JointImpedancePtr
     */
    static JointImpedancePtr create(const nemo::RealVector& values);

    /**
     * Get JointStiffness of all joints.
     *
     * @return Stiffness of all joints
     */
    JointStiffness getStiffness() const;

    /**
     *
     * @return
     */
    unsigned int getDimension() const;

    /**
     * Get JointStiffness of a single joint.
     *
     * @param index
     * 			Index of joint
     *
     * @return Stiffness of specified joint
     */
    JointStiffness getStiffness(unsigned int index) const;

    /**
     * Get JointDamping of all joints.
     *
     * @return Damping of all joints
     */
    JointDamping getDamping() const;

    /**
     * Get JointDamping of a single joint.
     *
     * @param index
     * 			Index of joint
     *
     * @return Damping of specified joint
     */
    JointDamping getDamping(unsigned int index) const;

    /**
     *
     *
     * @param index
     *
     * @return
     */
    JointImpedance operator[](unsigned int index) const;

    /**
     * Access impedance values as untyped doubles. Valid indices are twice the
     * impedance dimension, you get stiffness values for every even indices
     * (including zero) and damping values for all uneven indices.
     *
     * @param index
     * 			Index of value to get. You get stiffness values for every even
     * 			indices (including zero), and damping values for every even
     * 			indices.
     *
     * @return Stiffness (even index, including zero) or damping (uneven index)
     * 			value
     */
    virtual double asDouble(unsigned int index) const;

    /**
     * Access impedance values as untyped double vector. Size of the vector is
     * twice the size of the impedance DTO`s dimension. Even indices
     * (including zero) are stiffness value, uneven indices are damping values.
     *
     * @return Stiffness (even index, including zero) or damping (uneven index)
     * 			values in a vector.
     */
    virtual nemo::RealVector asDoubleVector() const;

    /**
     * Setting value at given index.
     *
     * @param index
     * 			Index of the value to set/ change. Even values (including zero)
     * 			set stiffness values, uneven values set damping values.
     * @param value
     * 			Value to set
     */
    virtual void setValue(unsigned int index, double value);

    /**
     * Merging several low-dimensional JointImpedance DTOs into one single
     * high-dimensional JointImpedance DTO.
     * Unfortunately this method has to have a template parameter to be
     * compatible to the generic method in ContainingDoubles.
     *
     * @tparam DTOType
     *          Type of input DTOs (automatically inferred from inputs if left
     *          blank)
     * @param inputDTOs
     *          Vector of JointImpedancePtr, that should be merged
     *          (in-place)
     */
    template<class DTOType>
    void merge(std::vector<boost::shared_ptr<DTOType> >& inputDTOs) {
        // Check dimensions of input DTOs and my own dimension
        unsigned int inputdimension = 0;
        for (unsigned int inputdto = 0; inputdto < inputDTOs.size();
                ++inputdto) {
            inputdimension += inputDTOs[inputdto]->getDimension();
        }
        if (inputdimension != this->getDimension()) {
            throw std::runtime_error(
                    "Merging failed: My own dimension and the dimension of the summarized input JointImpedance DTOs doesn`t agree.");
        }

        // Start merging
        unsigned int cursor = 0;
        nemo::RealVector dampings(nemo::dim(this->getDimension())), stiffnesses(
                nemo::dim(this->getDimension()));
        for (unsigned int inputdto = 0; inputdto < inputDTOs.size();
                ++inputdto) {
            for (unsigned int inputdim = 0;
                    inputdim < inputDTOs[inputdto]->getDimension();
                    ++inputdim) {
                this->_stiffness.setValue(cursor,
                        inputDTOs[inputdto]->asDouble(2 * inputdim));
                this->_damping.setValue(cursor++,
                        inputDTOs[inputdto]->asDouble(2 * inputdim + 1));
            }
        }
    }

    /**
     * Split this DTO into several one-dimensional DTOs.
     *
     * @tparam DTOType
     *          Type of input and return DTOs. Has to be JointImpedance but
     *          unfortunately has to be explicit to be compatible to the generic
     *          method in ContainingDoubles.
     *
     * @return Vector of one-dimensional JointImpedancePtr
     */
    template<class DTOType>
    std::vector<JointImpedancePtr> split() const {
        std::vector<JointImpedancePtr> vec;

        for (unsigned int cursor = 0; cursor < getDimension(); ++cursor) {
            nemo::RealVector values(nemo::dim(2), this->asDouble(2 * cursor));
            values[1] = this->asDouble(2 * cursor + 1);
            vec.push_back(JointImpedancePtr(new JointImpedance(values)));
        }

        return vec;
    }

protected:
    /**
     * Print function
     */
    virtual std::string print() const;

    /**
     * Stiffness value
     */
    JointStiffness _stiffness;

    /**
     * Stiffness value
     */
    JointDamping _damping;
};

}
