#pragma once

#include <iostream>
#include <sstream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/JointValues.h"

namespace rci {

class JointTorques;
typedef boost::shared_ptr<JointTorques> JointTorquesPtr;

/**
 * Domain object, representing Joint Torques.
 * Default representatin is newton meter, which is the SI measurement unit
 * of torque.
 * @ingroup DTO
 */
class JointTorques: public JointValues {

public:

    /**
     * Create JointTorques from value(s) in \f$N*m\f$
     * @param value
     * @return Joint torques
     */
    static JointTorquesPtr fromNm(double value);
    static JointTorquesPtr fromNm(const nemo::RealVector& values);

    /**
     * Empty constructor
     * Initializes JointTorques with no values.
     */
    JointTorques();

    /**
     * Constructor from double vector.
     * Initializes JointTorques with given values from double vector.
     *
     * @param values
     *          Double vector with torque values (\f$N*m\f$)
     */
    JointTorques(const nemo::RealVector& values);

    /**
     * Constructor from double vector.
     * Initializes JointTorques with given values from double vector.
     *
     * @param value
     *          Torque values (\f$N*m\f$)
     */
    JointTorques(double value);

    /**
     * Create blank JointTorquesPtr with given dimension (and optionally fill it
     * with the given value, otherwise filled with zeros).
     *
     * @param dimension
     *          Dimension of the JointTorques DTo to create
     * @param value
     *          Value to fill each end every field of the DTO (value in default
     *          representation - \fi$Nm\fi$)
     */
    static JointTorquesPtr create(unsigned int dimension, double value = 0.0);

    /**
     * Getter for the values of this Torques values in \fi$Nm\fi$
     *
     * @param index
     *          Index to get in \fi$Nm\fi$
     *
     * @return
     */
    virtual double Nm(unsigned int index) const;
    virtual nemo::RealVector NmVector() const;

    /**
     * Setter for the values of this Torque values in \fi$Nm\fi$
     *
     * @param index
     *          Index to set
     * @param value
     *          Value to set (in \fi$Nm\fi$)
     */
    virtual void setFromNm(unsigned int index, double value);
    virtual void setFromNm(const nemo::RealVector& values);

    /**
     * Create JointTorquesPtr from a JointTorques object (deep copy)
     *
     * @param JointTorques object
     */
    static JointTorquesPtr copy(const JointTorques&);

protected:
    /**
     * Print function
     */
    virtual std::string print() const;
};

}
