#pragma once

#include <iostream>
#include <sstream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/JointValues.h"

namespace rci {

class JointVelocities;
typedef boost::shared_ptr<JointVelocities> JointVelocitiesPtr;

/**
 * Domain object, representing Joint Velocities.
 * Default representation is in radian per second (\f$\frac{rad}{s}\f$),
 * according to the International System of Units (SI).
 * @ingroup DTO
 */
class JointVelocities: public JointValues {

public:
    JointVelocities();
    JointVelocities(nemo::RealVector values);
    JointVelocities(double value);

    /**
     * Creates one-dimensional JointVelocitiesPtr from a single double value.
     * You should not use this method, unless you exactly know what you are
     * doing. Try to rather use a more specific method.
     *
     * @param value
     *          Initial value.
     *
     * @return One-dimensional JointVelocitiesPtr
     */
    static JointVelocitiesPtr fromValue(double value);

    /**
     * Creates a JointVelocitiesPtr from a vector of double values.
     * You should not use this method, unless you exactly know what you are
     * doing. Try to rather use a more specific method.
     *
     * @param values
     *          Initial velocity values.
     *
     * @return JointVelocitiesPtr of the size of the given double vector.
     */
    static JointVelocitiesPtr fromValues(nemo::RealVector values);

    /**
     * Creates a one-dimensional JointVelocitiesPtr from a single \f$\frac{rad}{s}\f value.
     *
     * @param values
     *          Initial velocity value in \f$\frac{rad}{s}\f$
     *
     * @return One-dimensional JointVelocitiesPtr
     */
    static JointVelocitiesPtr fromRad_s(double value);

    /**
     * Creates a JointVelocitiesPtr from a vector of \f$\frac{rad}{s}\f values.
     *
     * @param values
     *          Initial velocity values in \f$\frac{rad}{s}\f$
     *
     * @return JointVelocitiesPtr of the size of the given vector.
     */
    static JointVelocitiesPtr fromRad_s(nemo::RealVector values);

    /**
     * Get all velocity values as double vector (in \f$\frac{rad}{s}\f$).
     *
     * @return Double vector (in \f$\frac{rad}{s}\f$)
     */
    virtual nemo::RealVector rad_sVector();

    /**
     * Get single velocity value in \f$\frac{rad}{s}\f.
     *
     * @param index
     *          Index of velocity value
     *
     * @return Single velocity value in \f$\frac{rad}{s}\f$
     */
    virtual double rad_s(unsigned int index);

    /**
     * Creates a one-dimensional JointVelocitiesPtr from a single
     * \f$\frac{deg}{s}\f$ value.
     *
     * @param values
     *          Initial velocity value in \f$\frac{deg}{s}\f$.
     *
     * @return One-dimensional JointVelocitiesPtr
     */
    static JointVelocitiesPtr fromDeg_s(double value);

    /**
     * Creates a JointVelocitiesPtr from a vector of \f$\frac{rad}{s}\f$ values.
     *
     * @param values
     *          Initial velocity values in \f$\frac{rad}{s}\f$.
     *
     * @return JointVelocitiesPtr of the size of the given vector.
     */
    static JointVelocitiesPtr fromDeg_s(nemo::RealVector values);

    /**
     * Get all velocity values as double vector (in \f$\frac{rad}{s}\f$).
     *
     * @return Double vector (in \f$\frac{rad}{s}\f$)
     */
    virtual nemo::RealVector deg_sVector();

    /**
     * Get single velocity value in \f$\frac{rad}{s}\f$.
     *
     * @param index
     *          Index of velocity value
     *
     * @return Single velocity value in \f$\frac{rad}{s}\f$
     */

    virtual double deg_s(unsigned int index);
    /**
     * Print function
     */
    virtual std::string print() const;
};

}
