#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/CartesianDamping.h"

using namespace nemo;
using namespace rci;

class CartesianDampingTest: public ::testing::Test {
protected:

    CartesianDampingTest() :
            values(), st() {
        values = RealVector(dim(7));
        values[0] = 1.12;
        values[1] = 2.0;
        values[2] = -1.1;
        values[3] = 5 * M_PI;
        values[4] = 0.0;
        values[5] = -3 * M_PI;
        values[6] = 0.0;
        st = CartesianDamping(values);
    }

    virtual ~CartesianDampingTest() {
    }

    RealVector values;
    CartesianDamping st;
};

TEST_F(CartesianDampingTest, testInitialization)
{
    EXPECT_EQ(7, CartesianDamping().getDimension());
}

TEST_F(CartesianDampingTest, testConstruction)
{
    // Check dimension
    CartesianDamping s = CartesianDamping(values);
    EXPECT_EQ(7, s.getDimension());

    // Check dimension
    s = CartesianDamping(1, 2, 3, 4, 5, 6,7);
    EXPECT_EQ(7, s.getDimension());
}

TEST_F(CartesianDampingTest, testTranslationalDamping)
{
    // Check dimension
    EXPECT_THROW(TranslationalDamping td = TranslationalDamping(values), std::invalid_argument);

    // Check dimension
    TranslationalDamping td = TranslationalDamping();
    EXPECT_EQ(3, td.getDimension());

    // Check dimension
    td = TranslationalDamping(1, 2, 3);
    EXPECT_EQ(3, td.getDimension());

    // Check getter
    EXPECT_DOUBLE_EQ(1, td.asDouble(0));
}

TEST_F(CartesianDampingTest, testRotationalDamping)
{
    // Check dimension
    EXPECT_THROW(RotationalDamping rd = RotationalDamping(values), std::invalid_argument);

    // Check dimension
    RotationalDamping rd = RotationalDamping();
    EXPECT_EQ(4, rd.getDimension());

    // Check dimension
    rd = RotationalDamping(1, 2, 3,4);
    EXPECT_EQ(4, rd.getDimension());

    // Check getter
    EXPECT_DOUBLE_EQ(1, rd.asDouble(0));
}

TEST_F(CartesianDampingTest, testValueGetter)
{
    RealVector testvalues;

    // Check example values
    EXPECT_DOUBLE_EQ(1.12, st.asDouble(0));
    EXPECT_DOUBLE_EQ(-1.1, st.asDouble(2));
    EXPECT_DOUBLE_EQ(0, st.asDouble(4));

    // Check values vector
    testvalues = st.asDoubleVector();
    EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
    EXPECT_DOUBLE_EQ(values[3], testvalues[3]);
    EXPECT_DOUBLE_EQ(values[5], testvalues[5]);
}

TEST_F(CartesianDampingTest, testCartesianGetter)
{
    // Check position values
    EXPECT_DOUBLE_EQ(1.12, st.x());
    EXPECT_DOUBLE_EQ(2.0, st.y());
    EXPECT_DOUBLE_EQ(-1.1, st.z());

    // Check orientation values
    EXPECT_DOUBLE_EQ(5 * M_PI, st.q0());
    EXPECT_DOUBLE_EQ(0.0, st.q1());
    EXPECT_DOUBLE_EQ(-3 * M_PI, st.q2());
    EXPECT_DOUBLE_EQ(0.0, st.q3());
}
