#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/CartesianPose.h"

static const double EPSILON = 10e-7;

using namespace nemo;
using namespace rci;

class OrientationTest: public ::testing::Test {
protected:

    OrientationTest() :
            values(), orientation() {
        values = RealVector(dim(4));
        values[0] = 0;
        values[1] = 1.0;
        values[2] = 0;
        values[3] = 0.0;
        orientation = Orientation(values);
    }

    virtual ~OrientationTest() {
    }

    RealVector values;
    Orientation orientation;
};

TEST_F(OrientationTest, testInitialization)
{
    EXPECT_EQ(4, Orientation().getDimension())
    ;}

TEST_F(OrientationTest, testConstruction)
{
    // Check dimension
    Orientation o = Orientation(values);
    EXPECT_EQ(4, o.getDimension())
    ;// Check dimension
    o = Orientation(1, 2, 3, 4);
    EXPECT_EQ(4, o.getDimension());
}

TEST_F(OrientationTest, testConstructionFromEulerAngles)
{
    OrientationPtr rotationX = Orientation::fromEulerAngles(M_PI, 0.0, 0.0);
    RealVector v1 = rotationX->asEulerAngles();
    EXPECT_NEAR(M_PI,v1[0], EPSILON)
    ; EXPECT_NEAR(0.0,v1[1], EPSILON);
    EXPECT_NEAR(0.0,v1[2], EPSILON);

    OrientationPtr rotationY = Orientation::fromEulerAngles(0.0, M_PI, 0.0);
    RealVector v2= rotationY->asEulerAngles();
    EXPECT_NEAR(M_PI,v2[0], EPSILON);
    EXPECT_NEAR(0.0,v2[1], EPSILON);
    EXPECT_NEAR(M_PI,v2[2], EPSILON);

    OrientationPtr rotationZ = Orientation::fromEulerAngles(0.0, 0.0, M_PI);
    RealVector v3= rotationZ->asEulerAngles();
    EXPECT_NEAR(0.0,v3[0], EPSILON);
    EXPECT_NEAR(0.0,v3[1], EPSILON);
    EXPECT_NEAR(M_PI,v3[2], EPSILON);

    OrientationPtr rotationX2 = Orientation::fromEulerAngles(2* M_PI, 0.0, 0.0);
    RealVector v4 = rotationX2->asEulerAngles();
    EXPECT_NEAR(0.0,v4[0], EPSILON);
    EXPECT_NEAR(0.0,v4[1], EPSILON);
    EXPECT_NEAR(0.0,v4[2], EPSILON);

    OrientationPtr rotationFoo = Orientation::fromEulerAngles(1.1, 2.2, 3.3);
    RealVector v5= rotationFoo->asEulerAngles();
    EXPECT_NEAR(-2.0415926535897926,v5[0], EPSILON);
    EXPECT_NEAR(0.94159265358979294,v5[1], EPSILON);
    EXPECT_NEAR(0.15840734641020662,v5[2], EPSILON);

    OrientationPtr rotationX3 = Orientation::fromEulerAngles(6* M_PI, 0.0, 0.0);
    RealVector v6 = rotationX3->asEulerAngles();
    EXPECT_NEAR(0.0,v6[0], EPSILON);
    EXPECT_NEAR(0.0,v6[1], EPSILON);
    EXPECT_NEAR(0.0,v6[2], EPSILON);

    OrientationPtr rotationFoo2 = Orientation::fromEulerAngles(5* M_PI, M_PI, 0.0);
    RealVector v7 = rotationFoo2->asEulerAngles();
    EXPECT_NEAR(0.0,v7[0], EPSILON);
    EXPECT_NEAR(0.0,v7[1], EPSILON);
    EXPECT_NEAR(M_PI,v7[2], EPSILON);
}

TEST_F(OrientationTest, testValueGetter)
{
    RealVector testvalues;

    // Check example values
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(0))
    ; EXPECT_DOUBLE_EQ(1, orientation.asDouble(1));
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(2));
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(3));

    // Check values vector
    testvalues = orientation.asDoubleVector();
    EXPECT_DOUBLE_EQ(values[0], testvalues[0]);
    EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
    EXPECT_DOUBLE_EQ(values[2], testvalues[2]);
    EXPECT_DOUBLE_EQ(values[3], testvalues[3]);
}

TEST_F(OrientationTest, testCartesianGetter)
{
    // Check position values
    EXPECT_DOUBLE_EQ(0, orientation.q0());
    EXPECT_DOUBLE_EQ(1.0, orientation.q1());
    EXPECT_DOUBLE_EQ(0, orientation.q2());
    EXPECT_DOUBLE_EQ(0, orientation.q3());
}

TEST_F(OrientationTest, testSetter)
{
    EXPECT_THROW(orientation.setValue(0.0), std::runtime_error);

    orientation.setValue(0, 2.0);
    orientation.setValue(1, 3.0);
    orientation.setValue(2, 4.0);
    orientation.setValue(3, 5.0);

    // Check position values
    EXPECT_DOUBLE_EQ(2.0, orientation.q0());
    EXPECT_DOUBLE_EQ(3.0, orientation.q1());
    EXPECT_DOUBLE_EQ(4.0, orientation.q2());
    EXPECT_DOUBLE_EQ(5.0, orientation.q3());
}

TEST_F(OrientationTest, testRotation)
{

    OrientationPtr rotationX = Orientation::fromEulerAngles(M_PI, 0.0, 0.0);
    OrientationPtr rotationY = Orientation::fromEulerAngles(0.0, M_PI, 0.0);
    OrientationPtr rotationZ = Orientation::fromEulerAngles(0.0, 0.0, M_PI);

    OrientationPtr rotationXY = rotationX->rotateBy(rotationY);
    RealVector vxy = rotationXY->asEulerAngles();
    EXPECT_NEAR(0.0, vxy[0], EPSILON)
    ; EXPECT_NEAR(0.0, vxy[1], EPSILON);
    EXPECT_NEAR(M_PI, vxy[2], EPSILON);

    OrientationPtr rotationYX = rotationY->rotateBy(rotationX);
    RealVector vyx = rotationYX->asEulerAngles();
    EXPECT_NEAR(0.0,vyx[0], EPSILON);
    EXPECT_NEAR(0.0,vyx[1], EPSILON);
    EXPECT_NEAR(M_PI,vyx[2], EPSILON);

    OrientationPtr rotationXX = rotationX->rotateBy(rotationX);
    RealVector vxx = rotationXX->asEulerAngles();
    EXPECT_NEAR(0.0,vxx[0], EPSILON);
    EXPECT_NEAR(0.0,vxx[1], EPSILON);
    EXPECT_NEAR(0.0,vxx[2], EPSILON);
}

TEST_F(OrientationTest, testCreationFromRotationMatrix)
{
    OrientationPtr RM;

    RM = Orientation::fromRotationMatrix(-1.0, 0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 0.0, -1.0);
    EXPECT_NEAR(0, RM->q0(), EPSILON);
    EXPECT_NEAR(0, RM->q1(), EPSILON);
    EXPECT_NEAR(0, RM->q2(), EPSILON);
    EXPECT_NEAR(-1, RM->q3(), EPSILON);

    RM = Orientation::fromRotationMatrix(0, -1, 0, -1, 0, 0, 0, 0, -1);
    EXPECT_NEAR(0, RM->q0(), EPSILON);
    EXPECT_NEAR(0.707107, RM->q1(), EPSILON);
    EXPECT_NEAR(-0.707107, RM->q2(), EPSILON);
    EXPECT_NEAR(0, RM->q3(), EPSILON);
}

TEST_F(OrientationTest, testCreationFromAxisAngle)
{
    OrientationPtr AA;

    AA = Orientation::fromAxisAngle(M_PI / 2, 1.0, 0.0, 0.0);
    EXPECT_NEAR(0.70710678, AA->q0(), EPSILON);
    EXPECT_NEAR(0.70710678, AA->q1(), EPSILON);
    EXPECT_NEAR(0, AA->q2(), EPSILON);
    EXPECT_NEAR(0, AA->q3(), EPSILON);
}

TEST_F(OrientationTest, testRotationAsAxisAngle)
{
    OrientationPtr r;
    RealVector aa;

    r = Orientation::fromQuaternion(sqrt(2) / 2, sqrt(2) / 2, 0.0, 0.0);
    aa = r->asAxisAngle();
    EXPECT_NEAR(M_PI / 2, aa[0], EPSILON)
    ; EXPECT_NEAR(1, aa[1], EPSILON);
    EXPECT_NEAR(0, aa[2], EPSILON);
    EXPECT_NEAR(0, aa[3], EPSILON);
}

TEST_F(OrientationTest, testEulerConversions)
{
    OrientationPtr r;
    RealVector veul = RealVector(dim(3));

    // Construction from quaternion
    r = Orientation::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through euler angles
    veul = r->asEulerAngles();
    r = Orientation::fromEulerAngles(veul[0], veul[1], veul[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);
}

TEST_F(OrientationTest, testAxisAngleConversions)
{
    OrientationPtr r;
    RealVector vaa = RealVector(dim(4));

    // Construction from quaternion
    r = Orientation::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vaa = r->asAxisAngle();
    r = Orientation::fromAxisAngle(vaa[0], vaa[1], vaa[2], vaa[3]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}

TEST_F(OrientationTest, testCompactAxisAngleConversions)
{
    OrientationPtr r;
    RealVector vaa = RealVector(dim(3));

    // Construction from quaternion
    r = Orientation::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Back and forth through axis-angle
    vaa = r->asCompactAxisAngle();
    r = Orientation::fromCompactAxisAngle(vaa[0], vaa[1], vaa[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}

TEST_F(OrientationTest, testChainOfAllConversions)
{
    OrientationPtr r;
    RealVector vaa = RealVector(dim(4));
    RealVector veul = RealVector(dim(3));
    RealVector vcaa = RealVector(dim(3));
    RealVector vrm = RealVector(dim(9));

    // Construction from quaternion
    r = Orientation::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through euler angles
    veul = r->asEulerAngles();
    r = Orientation::fromEulerAngles(veul[0], veul[1], veul[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vcaa = r->asCompactAxisAngle();
    r = Orientation::fromCompactAxisAngle(vcaa[0], vcaa[1], vcaa[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vaa = r->asAxisAngle();
    r = Orientation::fromAxisAngle(vaa[0], vaa[1], vaa[2], vaa[3]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}

TEST_F(OrientationTest, testRotationMatrixToEulerAngles)
{
    OrientationPtr RM;

    RM = Orientation::fromRotationMatrix(1, 0, 0, 0, 1, 0, 0, 0, 1);
    EXPECT_NEAR(0, RM->a(), EPSILON);
    EXPECT_NEAR(0, RM->b(), EPSILON);
    EXPECT_NEAR(0, RM->c(), EPSILON);

    RM = Orientation::fromRotationMatrix(-1, 0, 0, 0, 1, 0, 0, 0, -1);
    EXPECT_NEAR(M_PI, RM->a(), EPSILON);
    EXPECT_NEAR(0, RM->b(), EPSILON);
    EXPECT_NEAR(M_PI, RM->c(), EPSILON);

    RM = Orientation::fromRotationMatrix(0, 0, -1, 0, 1, 0, 1, 0, 0);
    EXPECT_NEAR(M_PI, RM->a(), EPSILON);
    EXPECT_NEAR(-M_PI/2, RM->b(), EPSILON);
    EXPECT_NEAR(M_PI, RM->c(), EPSILON);

    RM = Orientation::fromRotationMatrix(1, 0, 0, 0, 0, 1, 0, -1, 0);
    EXPECT_NEAR(-M_PI/2, RM->a(), EPSILON);
    EXPECT_NEAR(0, RM->b(), EPSILON);
    EXPECT_NEAR(0, RM->c(), EPSILON);

    RM = Orientation::fromRotationMatrix(0, -1, 0, -1, 0, 0, 0, 0, -1);
    EXPECT_NEAR(M_PI, RM->a(), EPSILON);
    EXPECT_NEAR(0, RM->b(), EPSILON);
    EXPECT_NEAR(-M_PI/2, RM->c(), EPSILON);
}

TEST_F(OrientationTest, testZeroRotation)
{
    OrientationPtr rot;
    RealVector vec;

    rot = Orientation::fromEulerAngles(0.0, 0.0, 0.0);

    EXPECT_DOUBLE_EQ(1.0, rot->q0());
    EXPECT_DOUBLE_EQ(0.0, rot->q1());
    EXPECT_DOUBLE_EQ(0.0, rot->q2());
    EXPECT_DOUBLE_EQ(0.0, rot->q3());

    EXPECT_DOUBLE_EQ(0.0, rot->a());
    EXPECT_DOUBLE_EQ(0.0, rot->b());
    EXPECT_DOUBLE_EQ(0.0, rot->c());

    vec = rot->asAxisAngle();
    EXPECT_DOUBLE_EQ(0.0, vec[0]);
    EXPECT_DOUBLE_EQ(0.0, vec[1]);
    EXPECT_DOUBLE_EQ(0.0, vec[2]);
    EXPECT_DOUBLE_EQ(0.0, vec[3]);

    vec = rot->asCompactAxisAngle();
    EXPECT_DOUBLE_EQ(0.0, vec[0]);
    EXPECT_DOUBLE_EQ(0.0, vec[1]);
    EXPECT_DOUBLE_EQ(0.0, vec[2]);
}
