#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/CartesianPose.h"

using namespace std;
using namespace nemo;
using namespace rci;

static const double EPSILON = 10e-7;

class PoseTest: public ::testing::Test {
protected:

	PoseTest() :
			values(), pose() {
		values = RealVector(dim(7));
		values[0] = 1.12;
		values[1] = 2.0;
		values[2] = -1.1;
		values[3] = 5 * M_PI;
		values[4] = 0.0;
		values[5] = -3 * M_PI;
		values[6] = 0.0;
		pose = Pose(values);
	}

	virtual ~PoseTest() {
	}

	RealVector values;
	Pose pose;
};

TEST_F(PoseTest, testConstructionFromDoubleValues) {
	// Check dimension
	Pose p = Pose(1, 2, 3, 4, 5, 6, 7);
	EXPECT_EQ(7, p.getDimension());
	EXPECT_DOUBLE_EQ(1, p.asDouble(0));
	EXPECT_DOUBLE_EQ(3, p.asDouble(2));
	EXPECT_DOUBLE_EQ(4, p.asDouble(3));
	EXPECT_DOUBLE_EQ(7, p.asDouble(6));
}

TEST_F(PoseTest, testConstructionFromDoubleVector) {
	// Check dimension
	Pose p = Pose(values);
	EXPECT_EQ(7, p.getDimension());
	EXPECT_DOUBLE_EQ(1.12, p.asDouble(0));
	EXPECT_DOUBLE_EQ(-1.1, p.asDouble(2));
	EXPECT_DOUBLE_EQ(5 * M_PI, p.asDouble(3));
	EXPECT_DOUBLE_EQ(0.0, p.asDouble(6));
}

TEST_F(PoseTest, testConstructionFromTranslationAndRotation) {
	TranslationPtr t = Translation::fromMeters(1, 2, 3);
	OrientationPtr r = Orientation::fromQuaternion(0.1, 0.2, 0.3, 0.4);

	// Check dimension
	Pose p = Pose(t, r);
	EXPECT_EQ(7, p.getDimension());
	EXPECT_DOUBLE_EQ(1, p.asDouble(0));
	EXPECT_DOUBLE_EQ(3, p.asDouble(2));
	EXPECT_DOUBLE_EQ(0.1, p.asDouble(3));
	EXPECT_DOUBLE_EQ(0.4, p.asDouble(6));
}

TEST_F(PoseTest, testCopyConstructor) {

	const Pose p = Pose(values);
	Pose copy = p;

	EXPECT_EQ(p.getDimension(), copy.getDimension());
	EXPECT_DOUBLE_EQ(p.asDouble(0), copy.asDouble(0));
	EXPECT_DOUBLE_EQ(p.asDouble(2), copy.asDouble(2));
	EXPECT_DOUBLE_EQ(p.asDouble(3), copy.asDouble(3));
	EXPECT_DOUBLE_EQ(p.asDouble(6), copy.asDouble(6));
}

TEST_F(PoseTest, testValueGetter) {
	RealVector testvalues;

	// Check example values
	EXPECT_DOUBLE_EQ(1.12, pose.asDouble(0));
	EXPECT_DOUBLE_EQ(-1.1, pose.asDouble(2));
	EXPECT_DOUBLE_EQ(0, pose.asDouble(4));

	// Check values vector
	testvalues = pose.asDoubleVector();
	EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
	EXPECT_DOUBLE_EQ(values[3], testvalues[3]);
	EXPECT_DOUBLE_EQ(values[5], testvalues[5]);
}

TEST_F(PoseTest, testCartesianGetter) {
	// Check position values
	EXPECT_DOUBLE_EQ(1.12, pose.x());
	EXPECT_DOUBLE_EQ(2.0, pose.y());
	EXPECT_DOUBLE_EQ(-1.1, pose.z());

	// Check orientation values
	EXPECT_DOUBLE_EQ(5 * M_PI, pose.q0());
	EXPECT_DOUBLE_EQ(0.0, pose.q1());
	EXPECT_DOUBLE_EQ(-3 * M_PI, pose.q2());
	EXPECT_DOUBLE_EQ(0.0, pose.q3());
}

TEST_F(PoseTest, testSetter) {
	pose.setValue(0, 0.0);
	pose.setValue(2, 0.0);

	// Check position values
	EXPECT_DOUBLE_EQ(0.0, pose.x());
	EXPECT_DOUBLE_EQ(0.0, pose.z());

	pose.setValue(3, 0.0);
	pose.setValue(5, 0.0);

	// Check orientation values
	EXPECT_DOUBLE_EQ(0.0, pose.q0());
	EXPECT_DOUBLE_EQ(0.0, pose.q2());
}

TEST_F(PoseTest, testGetterNormalized) {
	RealVector testvalues;

	// Check example values
	EXPECT_DOUBLE_EQ(1.12, pose.asDouble(0));
	// No change
	EXPECT_DOUBLE_EQ(2.0, pose.asDouble(1));
	// No change
	EXPECT_DOUBLE_EQ(-1.1, pose.asDouble(2));
	// No change
	EXPECT_DOUBLE_EQ(5 * M_PI, pose.asDouble(3));
	// Normalization
	EXPECT_DOUBLE_EQ(0, pose.asDouble(4));
	// No change
	EXPECT_DOUBLE_EQ(-3 * M_PI, pose.asDouble(5));
	// Normalization
	EXPECT_DOUBLE_EQ(0, pose.asDouble(6));
	// Normalization

	// Check values vector
	testvalues = pose.asDoubleVector();
	EXPECT_DOUBLE_EQ(values[0], testvalues[0]);
	// No change
	EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
	// No change
	EXPECT_DOUBLE_EQ(values[2], testvalues[2]);
	// No change
	EXPECT_DOUBLE_EQ(5 * M_PI, testvalues[3]);
	// Normalization
	EXPECT_DOUBLE_EQ(values[4], testvalues[4]);
	// No change
	EXPECT_DOUBLE_EQ(-3 * M_PI, testvalues[5]);
	// Normalization
	EXPECT_DOUBLE_EQ(values[6], testvalues[6]);
	// No change
}

TEST_F(PoseTest, testFromMetersAndRadiansHomogeneous) {
	RealMatrix homogeneous(dim(3, 4), 0.0);
	homogeneous(0, 0) = 0.0;
	homogeneous(0, 1) = -1.0;
	homogeneous(0, 2) = 0.0;
	homogeneous(1, 0) = -1.0;
	homogeneous(1, 1) = 0.0;
	homogeneous(1, 2) = 0.0;
	homogeneous(2, 0) = 0.0;
	homogeneous(2, 1) = 0.0;
	homogeneous(2, 2) = -1.0;

	homogeneous(0, 3) = 1.11;
	homogeneous(1, 3) = 2.22;
	homogeneous(2, 3) = 3.33;

	PosePtr p = Pose::fromMetersAndRadiansHomogeneous(homogeneous);

	RealVector testvalues = p->asDoubleVector();

	EXPECT_NEAR(1.11, testvalues[0], EPSILON);
	EXPECT_NEAR(2.22, testvalues[1], EPSILON);
	EXPECT_NEAR(3.33, testvalues[2], EPSILON);

	EXPECT_NEAR(0.0, testvalues[3], EPSILON);
	EXPECT_NEAR(0.707107, testvalues[4], EPSILON);
	EXPECT_NEAR(-0.707107, testvalues[5], EPSILON);
	EXPECT_NEAR(0.0, testvalues[6], EPSILON);
}

TEST_F(PoseTest, regressionPoseFromTranslationAndOrientation) {
	PosePtr p = PosePtr(
			new Pose(Translation::fromMeters(0.55, 0, 0.25),
					Orientation::fromEulerAngles(0.0, M_PI, 0.0)));
	EXPECT_EQ(0.55, p->_transl.asDouble(0));
}

TEST_F(PoseTest, regressionRotationMatrixToCompactAxisAngles) {
	PosePtr ppose = Pose::fromMetersAndRadiansHomogeneous(-0.9996185303,
			-0.02710843459, 0.005286611617, -0.4500545561, -0.02711678855,
			0.9269508123, -0.3742015958, -0.5852395296, 0.005243589636,
			-0.3742021918, -0.9273323417, 0.1688518077);

	// Test if nan
	EXPECT_EQ(ppose->a(), ppose->a());
	EXPECT_EQ(ppose->b(), ppose->b());
	EXPECT_EQ(ppose->c(), ppose->c());
}

TEST_F(PoseTest, regressionRotationMatrixToCompactAxisAngles2) {
	PosePtr ppose = Pose::fromMetersAndRadiansHomogeneous(-0.9984720349,
			-0.0552579537, 0.0004264877061, -0.5783057809, -0.0552588217,
			0.9983872175, -0.01301411726, -0.005522137508, 0.0002933336364,
			-0.01301779971, -0.9999152422, 0.1988631487);

	// Test if nan
	EXPECT_EQ(ppose->a(), ppose->a());
	EXPECT_EQ(ppose->b(), ppose->b());
	EXPECT_EQ(ppose->c(), ppose->c());
}
