#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/JointAngles.h"

using namespace nemo;
using namespace std;
using namespace rci;

class JointAnglesTest: public ::testing::Test {
protected:

	JointAnglesTest() :
			values(), ja() {
		values = RealVector(dim(3));
		values[0] = 1.2 * M_PI;
		values[1] = -3 * M_PI;
		values[2] = -2 * M_PI;
		ja = JointAngles(values);
	}

	virtual ~JointAnglesTest() {
	}

	RealVector values;
	JointAngles ja;
};

TEST_F(JointAnglesTest, isEmptyInitially) {
	EXPECT_EQ(0, JointAngles().getDimension());
}

TEST_F(JointAnglesTest, testConstruction) {
	// Check dimension
	JointAngles j = JointAngles(values);
	EXPECT_EQ(3, j.getDimension());
}

TEST_F(JointAnglesTest, testConstructionFromSingleDouble) {
	// Check dimension
	JointAngles j = JointAngles(1.234);

	EXPECT_EQ(1, j.getDimension());
	EXPECT_DOUBLE_EQ(1.234, j.asDouble(0));
}

TEST_F(JointAnglesTest, testCopyContructor) {
	JointAngles cp(ja);
	EXPECT_DOUBLE_EQ(ja.asDouble(0), cp.asDouble(0));
	EXPECT_DOUBLE_EQ(ja.asDouble(1), cp.asDouble(1));
	EXPECT_DOUBLE_EQ(ja.asDouble(2), cp.asDouble(2));
}

TEST_F(JointAnglesTest, testStaticCopyCreation) {
	JointAnglesPtr jaPtr = JointAngles::copy(ja);

	EXPECT_EQ(ja.getDimension(), jaPtr->getDimension());
	EXPECT_DOUBLE_EQ(ja.asDouble(0), jaPtr->asDouble(0));
	EXPECT_DOUBLE_EQ(ja.asDouble(1), jaPtr->asDouble(1));
	EXPECT_DOUBLE_EQ(ja.asDouble(2), jaPtr->asDouble(2));
}

TEST_F(JointAnglesTest, testPointerCopy) {
	JointAnglesPtr jaPtr = JointAnglesPtr(new JointAngles(values));
	JointAnglesPtr ptrCpy = jaPtr;

	EXPECT_EQ(jaPtr->getDimension(), ptrCpy->getDimension());
	EXPECT_DOUBLE_EQ(jaPtr->asDouble(0), ptrCpy->asDouble(0));
	EXPECT_DOUBLE_EQ(jaPtr->asDouble(1), ptrCpy->asDouble(1));
	EXPECT_DOUBLE_EQ(jaPtr->asDouble(2), ptrCpy->asDouble(2));
}

TEST_F(JointAnglesTest, testGetters) {
	RealVector testvalues;

	// Check example values
	EXPECT_DOUBLE_EQ(1.2 * M_PI, ja.asDouble(0));
	EXPECT_DOUBLE_EQ(-3 * M_PI, ja.asDouble(1));
	EXPECT_DOUBLE_EQ(-2 * M_PI, ja.asDouble(2));

	// Check values vector
	testvalues = ja.asDoubleVector();
	EXPECT_DOUBLE_EQ(values[0], testvalues[0]);
	EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
	EXPECT_DOUBLE_EQ(values[2], testvalues[2]);
}

TEST_F(JointAnglesTest, testNormalizedGetters) {
	RealVector testvalues;

	// Check example values
	EXPECT_DOUBLE_EQ(1.2 * M_PI, ja.asDouble(0, true));
	// Not cropped
	EXPECT_DOUBLE_EQ(M_PI, ja.asDouble(1, true));
	// Cropped to [0, 2*pi[
	EXPECT_DOUBLE_EQ(0, ja.asDouble(2, true));
	// Cropped to [0, 2*pi[

	// Check values vector
	testvalues = ja.asDoubleVector(true);
	EXPECT_DOUBLE_EQ(values[0], testvalues[0]);
	// Not cropped
	EXPECT_DOUBLE_EQ(M_PI, testvalues[1]);
	// Cropped to [0, 2*pi[
	EXPECT_DOUBLE_EQ(0, testvalues[2]);
	// Cropped to [0, 2*pi[
}

TEST_F(JointAnglesTest, testDegToRad) {
	RealVector degrees(dim(3)), radians(dim(3));

	degrees[0] = 0;
	degrees[1] = -90;
	degrees[2] = 270;

	radians[0] = 0;
	radians[1] = -M_PI / 2;
	radians[2] = 3 * M_PI / 2;

	// Degree to radians
	JointAnglesPtr jad = JointAngles::fromDeg(degrees);
	EXPECT_DOUBLE_EQ(radians[0], jad->rad());
	EXPECT_DOUBLE_EQ(radians[1], jad->rad(1));
	EXPECT_DOUBLE_EQ(radians[2], jad->rad(2));

	// Radians to degrees
	JointAnglesPtr jar = JointAngles::fromRad(radians);
	EXPECT_DOUBLE_EQ(degrees[0], jar->deg());
	EXPECT_DOUBLE_EQ(degrees[1], jar->deg(1));
	EXPECT_DOUBLE_EQ(degrees[2], jar->deg(2));
}

TEST_F(JointAnglesTest, testDegreeVector) {
	RealVector degrees(dim(3)), radians(dim(3));

	radians[0] = 0;
	radians[1] = -M_PI / 2;
	radians[2] = 3 * M_PI / 2;

	degrees[0] = 0;
	degrees[1] = -90;
	degrees[2] = 270;

	// Degree to radians
	JointAnglesPtr jad = JointAngles::fromDeg(degrees);
	EXPECT_DOUBLE_EQ(degrees[0], jad->degVector()[0]);
	EXPECT_DOUBLE_EQ(degrees[1], jad->degVector()[1]);
	EXPECT_DOUBLE_EQ(degrees[2], jad->degVector()[2]);

	// Degree to radians
	EXPECT_DOUBLE_EQ(radians[0], jad->radVector()[0]);
	EXPECT_DOUBLE_EQ(radians[1], jad->radVector()[1]);
	EXPECT_DOUBLE_EQ(radians[2], jad->radVector()[2]);
}

TEST_F(JointAnglesTest, testRadiansVector) {
	RealVector degrees(dim(3)), radians(dim(3));

	radians[0] = 0;
	radians[1] = -M_PI / 2;
	radians[2] = 3 * M_PI / 2;

	degrees[0] = 0;
	degrees[1] = -90;
	degrees[2] = 270;

	// Degree to radians
	JointAnglesPtr jad = JointAngles::fromRad(radians);
	EXPECT_DOUBLE_EQ(radians[0], jad->radVector()[0]);
	EXPECT_DOUBLE_EQ(radians[1], jad->radVector()[1]);
	EXPECT_DOUBLE_EQ(radians[2], jad->radVector()[2]);

	// Degree to radians
	EXPECT_DOUBLE_EQ(degrees[0], jad->degVector()[0]);
	EXPECT_DOUBLE_EQ(degrees[1], jad->degVector()[1]);
	EXPECT_DOUBLE_EQ(degrees[2], jad->degVector()[2]);
}

TEST_F(JointAnglesTest, testSetters) {
	RealVector degrees(dim(3)), radians(dim(3));

	radians[0] = 0;
	radians[1] = -M_PI / 2;
	radians[2] = 3 * M_PI / 2;

	degrees[0] = 0;
	degrees[1] = -90;
	degrees[2] = 270;

	// test dimension mismatch exception
	JointAnglesPtr wrongSize = JointAngles::fromRad(RealVector(dim(2), 0.0));
	EXPECT_THROW(wrongSize->setValues(degrees), std::runtime_error);
	EXPECT_THROW(wrongSize->setValue(2, 180.0), std::runtime_error);
	EXPECT_THROW(wrongSize->setFromDeg(degrees), std::runtime_error);
	EXPECT_THROW(wrongSize->setFromDeg(2, 180.0), std::runtime_error);
	EXPECT_THROW(wrongSize->setFromRad(radians), std::runtime_error);
	EXPECT_THROW(wrongSize->setFromRad(2, M_PI), std::runtime_error);

	// test values that are set
	JointAnglesPtr angles = JointAngles::fromRad(RealVector(dim(3), 0.0));
	angles->setValue(1, radians[1]);
	EXPECT_DOUBLE_EQ(radians[1], angles->rad(1));
	angles->setValues(radians);
	EXPECT_DOUBLE_EQ(radians[1], angles->rad(1));

	angles->setFromRad(1, radians[1]);
	EXPECT_DOUBLE_EQ(radians[1], angles->rad(1));
	angles->setFromRad(radians);
	EXPECT_DOUBLE_EQ(radians[1], angles->rad(1));

	angles->setFromDeg(1, degrees[1]);
	EXPECT_DOUBLE_EQ(degrees[1], angles->deg(1));
	angles->setFromDeg(degrees);
	EXPECT_DOUBLE_EQ(degrees[1], angles->deg(1));
}
