#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/dto/CartesianVelocity.h"

static const double EPSILON = 10e-7;

using namespace nemo;
using namespace rci;

class RotationalVelocityTest: public ::testing::Test {
protected:

    RotationalVelocityTest() :
            values(), orientation() {
        values = RealVector(dim(4));
        values[0] = 0;
        values[1] = 1.0;
        values[2] = 0;
        values[3] = 0.0;
        orientation = RotationalVelocity(values);
    }

    virtual ~RotationalVelocityTest() {
    }

    RealVector values;
    RotationalVelocity orientation;
};

TEST_F(RotationalVelocityTest, testInitialization)
{
    EXPECT_EQ(4, RotationalVelocity().getDimension())
    ;}

TEST_F(RotationalVelocityTest, testConstruction)
{
    // Check dimension
    RotationalVelocity o = RotationalVelocity(values);
    EXPECT_EQ(4, o.getDimension())
    ;// Check dimension
    o = RotationalVelocity(1, 2, 3, 4);
    EXPECT_EQ(4, o.getDimension());
}

TEST_F(RotationalVelocityTest, testConstructionFromEulerAngles)
{
    RotationalVelocityPtr rotationX = RotationalVelocity::fromEulerAngles(M_PI, 0.0, 0.0);
    RealVector v1 = rotationX->asEulerAngles();
    EXPECT_NEAR(M_PI,v1[0], EPSILON)
    ; EXPECT_NEAR(0.0,v1[1], EPSILON);
    EXPECT_NEAR(0.0,v1[2], EPSILON);

    RotationalVelocityPtr rotationY = RotationalVelocity::fromEulerAngles(0.0, M_PI, 0.0);
    RealVector v2= rotationY->asEulerAngles();
    EXPECT_NEAR(M_PI,v2[0], EPSILON);
    EXPECT_NEAR(0.0,v2[1], EPSILON);
    EXPECT_NEAR(M_PI,v2[2], EPSILON);

    RotationalVelocityPtr rotationZ = RotationalVelocity::fromEulerAngles(0.0, 0.0, M_PI);
    RealVector v3= rotationZ->asEulerAngles();
    EXPECT_NEAR(0.0,v3[0], EPSILON);
    EXPECT_NEAR(0.0,v3[1], EPSILON);
    EXPECT_NEAR(M_PI,v3[2], EPSILON);

    RotationalVelocityPtr rotationX2 = RotationalVelocity::fromEulerAngles(2* M_PI, 0.0, 0.0);
    RealVector v4 = rotationX2->asEulerAngles();
    EXPECT_NEAR(0.0,v4[0], EPSILON);
    EXPECT_NEAR(0.0,v4[1], EPSILON);
    EXPECT_NEAR(0.0,v4[2], EPSILON);

    RotationalVelocityPtr rotationFoo = RotationalVelocity::fromEulerAngles(1.1, 2.2, 3.3);
    RealVector v5= rotationFoo->asEulerAngles();
    EXPECT_NEAR(-2.0415926535897926,v5[0], EPSILON);
    EXPECT_NEAR(0.94159265358979294,v5[1], EPSILON);
    EXPECT_NEAR(0.15840734641020662,v5[2], EPSILON);

    RotationalVelocityPtr rotationX3 = RotationalVelocity::fromEulerAngles(6* M_PI, 0.0, 0.0);
    RealVector v6 = rotationX3->asEulerAngles();
    EXPECT_NEAR(0.0,v6[0], EPSILON);
    EXPECT_NEAR(0.0,v6[1], EPSILON);
    EXPECT_NEAR(0.0,v6[2], EPSILON);

    RotationalVelocityPtr rotationFoo2 = RotationalVelocity::fromEulerAngles(5* M_PI, M_PI, 0.0);
    RealVector v7 = rotationFoo2->asEulerAngles();
    EXPECT_NEAR(0.0,v7[0], EPSILON);
    EXPECT_NEAR(0.0,v7[1], EPSILON);
    EXPECT_NEAR(M_PI,v7[2], EPSILON);
}

TEST_F(RotationalVelocityTest, testValueGetter)
{
    RealVector testvalues;

    // Check example values
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(0))
    ; EXPECT_DOUBLE_EQ(1, orientation.asDouble(1));
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(2));
    EXPECT_DOUBLE_EQ(0, orientation.asDouble(3));

    // Check values vector
    testvalues = orientation.asDoubleVector();
    EXPECT_DOUBLE_EQ(values[0], testvalues[0]);
    EXPECT_DOUBLE_EQ(values[1], testvalues[1]);
    EXPECT_DOUBLE_EQ(values[2], testvalues[2]);
    EXPECT_DOUBLE_EQ(values[3], testvalues[3]);
}

TEST_F(RotationalVelocityTest, testCartesianGetter)
{
    // Check position values
    EXPECT_DOUBLE_EQ(0, orientation.q0())
    ; EXPECT_DOUBLE_EQ(1.0, orientation.q1());
    EXPECT_DOUBLE_EQ(0, orientation.q2());
    EXPECT_DOUBLE_EQ(0, orientation.q3());
}

TEST_F(RotationalVelocityTest, testCreationFromAxisAngle)
{
    RotationalVelocityPtr AA;

    AA = RotationalVelocity::fromAxisAngle(M_PI / 2, 1.0, 0.0, 0.0);
    EXPECT_NEAR(0.70710678, AA->q0(), EPSILON)
    ; EXPECT_NEAR(0.70710678, AA->q1(), EPSILON);
    EXPECT_NEAR(0, AA->q2(), EPSILON);
    EXPECT_NEAR(0, AA->q3(), EPSILON);
}

TEST_F(RotationalVelocityTest, testRotationAsAxisAngle)
{
    RotationalVelocityPtr r;
    RealVector aa;

    r = RotationalVelocity::fromQuaternion(sqrt(2) / 2, sqrt(2) / 2, 0.0, 0.0);
    aa = r->asAxisAngle();
    EXPECT_NEAR(M_PI / 2, aa[0], EPSILON)
    ; EXPECT_NEAR(1, aa[1], EPSILON);
    EXPECT_NEAR(0, aa[2], EPSILON);
    EXPECT_NEAR(0, aa[3], EPSILON);
}

TEST_F(RotationalVelocityTest, testEulerConversions)
{
    RotationalVelocityPtr r;
    RealVector veul = RealVector(dim(3));

    // Construction from quaternion
    r = RotationalVelocity::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through euler angles
    veul = r->asEulerAngles();
    r = RotationalVelocity::fromEulerAngles(veul[0], veul[1], veul[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);
}

TEST_F(RotationalVelocityTest, testAxisAngleConversions)
{
    RotationalVelocityPtr r;
    RealVector vaa = RealVector(dim(4));

    // Construction from quaternion
    r = RotationalVelocity::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vaa = r->asAxisAngle();
    r = RotationalVelocity::fromAxisAngle(vaa[0], vaa[1], vaa[2], vaa[3]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}

TEST_F(RotationalVelocityTest, testCompactAxisAngleConversions)
{
    RotationalVelocityPtr r;
    RealVector vaa = RealVector(dim(3));

    // Construction from quaternion
    r = RotationalVelocity::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Back and forth through axis-angle
    vaa = r->asCompactAxisAngle();
    r = RotationalVelocity::fromCompactAxisAngle(vaa[0], vaa[1], vaa[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}

TEST_F(RotationalVelocityTest, testChainOfAllConversions)
{
    RotationalVelocityPtr r;
    RealVector vaa = RealVector(dim(4));
    RealVector veul = RealVector(dim(3));
    RealVector vcaa = RealVector(dim(3));
    RealVector vrm = RealVector(dim(9));

    // Construction from quaternion
    r = RotationalVelocity::fromQuaternion(sqrt(2) / 2, 0.0, sqrt(2) / 2, 0.0);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through euler angles
    veul = r->asEulerAngles();
    r = RotationalVelocity::fromEulerAngles(veul[0], veul[1], veul[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vcaa = r->asCompactAxisAngle();
    r = RotationalVelocity::fromCompactAxisAngle(vcaa[0], vcaa[1], vcaa[2]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

    // Back and forth through axis-angle
    vaa = r->asAxisAngle();
    r = RotationalVelocity::fromAxisAngle(vaa[0], vaa[1], vaa[2], vaa[3]);

    // Check
    EXPECT_NEAR(0.70710678, r->q0(), EPSILON)
    ; EXPECT_NEAR(0, r->q1(), EPSILON);
    EXPECT_NEAR(0.70710678, r->q2(), EPSILON);
    EXPECT_NEAR(0, r->q3(), EPSILON);

}
