#include <gtest/gtest.h>

#include <iostream>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include "rci/ResourceNode.h"
#include "rci/Sensing.h"

using namespace nemo;
using namespace rci;

class CartesianNode: public ResourceNode,
        public Sensing,
        public PoseSensing,
        public CartesianVelocitySensing,
        public CartesianAccelerationSensing {
public:
    CartesianNode() :
            ResourceNode("Cartesian Sensing Node"), Sensing(), PoseSensing(), CartesianVelocitySensing(), CartesianAccelerationSensing() {
    }
    virtual ~CartesianNode() {
    }
};

class CartesianSensingTest: public ::testing::Test {
protected:

    CartesianSensingTest() :
            cart(), pose(), orientation(), translation(), velocity(), transvel(), rotvel(), acceleration(), transacc(), rotacc() {

        translation = Translation::fromMeters(1.0, 2.0, 3.0);
        orientation = Orientation::fromEulerAngles(M_PI, 0.0, M_PI);
        pose = PosePtr(new Pose(translation, orientation));

        transvel = TranslationalVelocity::fromM_s(1.0, 2.0, 3.0);
        rotvel = RotationalVelocity::fromEulerAngles(M_PI, 0.0, M_PI);
        velocity = VelocityPtr();

        transacc = TranslationalAcceleration::fromM_s2(1.0, 2.0, 3.0);
        rotacc = RotationalAccelerationPtr();
        acceleration = AccelerationPtr();
    }

    virtual ~CartesianSensingTest() {
    }

    CartesianNode cart;

    PosePtr pose;
    OrientationPtr orientation;
    TranslationPtr translation;

    VelocityPtr velocity;
    TranslationalVelocityPtr transvel;
    RotationalVelocityPtr rotvel;

    AccelerationPtr acceleration;
    TranslationalAccelerationPtr transacc;
    RotationalAccelerationPtr rotacc;
};

TEST_F(CartesianSensingTest, testUpdateTranslation)
{
    EXPECT_THROW(cart.getTranslation(), std::runtime_error);

    cart.updateTranslation(translation);

    // Test pose
    EXPECT_THROW(cart.getPose(), std::runtime_error);

    // Test translation
    EXPECT_DOUBLE_EQ(1.0, cart.getTranslation()->x());
    EXPECT_DOUBLE_EQ(2.0, cart.getTranslation()->y());
    EXPECT_DOUBLE_EQ(3.0, cart.getTranslation()->z());

    // Test orientation
    EXPECT_THROW(cart.getOrientation(), std::runtime_error);
}

TEST_F(CartesianSensingTest, testUpdateOrientation)
{
    EXPECT_THROW(cart.getOrientation(), std::runtime_error);

    cart.updateOrientation(orientation);

    // Test pose
    EXPECT_THROW(cart.getPose(), std::runtime_error);

    // Test translation
    EXPECT_THROW(cart.getTranslation(), std::runtime_error);

    // Test orientation
    EXPECT_DOUBLE_EQ(M_PI, cart.getOrientation()->a());
    EXPECT_DOUBLE_EQ(0.0, cart.getOrientation()->b());
    EXPECT_DOUBLE_EQ(M_PI, cart.getOrientation()->c());
}

TEST_F(CartesianSensingTest, testUpdatePose)
{
    cart.updatePose(pose);

    // Test pose
    EXPECT_DOUBLE_EQ(1.0, cart.getPose()->x());
    EXPECT_DOUBLE_EQ(2.0, cart.getPose()->y());
    EXPECT_DOUBLE_EQ(3.0, cart.getPose()->z());
    EXPECT_DOUBLE_EQ(M_PI, cart.getPose()->a());
    EXPECT_DOUBLE_EQ(0.0, cart.getPose()->b());
    EXPECT_DOUBLE_EQ(M_PI, cart.getPose()->c());

    // Test translation
    EXPECT_DOUBLE_EQ(1.0, cart.getPose()->getTranslation()->x());
    EXPECT_DOUBLE_EQ(2.0, cart.getPose()->getTranslation()->y());
    EXPECT_DOUBLE_EQ(3.0, cart.getPose()->getTranslation()->z());

    // Test orientation
    EXPECT_DOUBLE_EQ(M_PI, cart.getPose()->getOrientation()->a());
    EXPECT_DOUBLE_EQ(0.0, cart.getPose()->getOrientation()->b());
    EXPECT_DOUBLE_EQ(M_PI, cart.getPose()->getOrientation()->c());
}

TEST_F(CartesianSensingTest, testUpdateTranslationalVelocity)
{
    EXPECT_THROW(cart.getTranslationalVelocity(), std::runtime_error);

    cart.updateTranslationalVelocity(transvel);

    // Test pose
    EXPECT_THROW(cart.getVelocity(), std::runtime_error);

    // Test translation
    EXPECT_DOUBLE_EQ(1.0, cart.getTranslationalVelocity()->x());
    EXPECT_DOUBLE_EQ(2.0, cart.getTranslationalVelocity()->y());
    EXPECT_DOUBLE_EQ(3.0, cart.getTranslationalVelocity()->z());

    // Test orientation
    EXPECT_THROW(cart.getRotationalVelocity(), std::runtime_error);
}

TEST_F(CartesianSensingTest, testUpdateRotationalVelocity)
{
    EXPECT_THROW(cart.getRotationalVelocity(), std::runtime_error);

    cart.updateRotationalVelocity(rotvel);

    // Test pose
    EXPECT_THROW(cart.getVelocity(), std::runtime_error);

    // Test translation
    EXPECT_THROW(cart.getTranslationalVelocity(), std::runtime_error);

    // Test orientation
    EXPECT_DOUBLE_EQ(M_PI, cart.getRotationalVelocity()->a());
    EXPECT_DOUBLE_EQ(0.0, cart.getRotationalVelocity()->b());
    EXPECT_DOUBLE_EQ(M_PI, cart.getRotationalVelocity()->c());
}

TEST_F(CartesianSensingTest, testUpdateTranslationalAcceleration)
{
    EXPECT_THROW(cart.getTranslationalAcceleration(), std::runtime_error);

    cart.updateTranslationalAcceleration(transacc);

    // Test pose
    EXPECT_THROW(cart.getAcceleration(), std::runtime_error);

    // Test translation
    EXPECT_DOUBLE_EQ(1.0, cart.getTranslationalAcceleration()->x());
    EXPECT_DOUBLE_EQ(2.0, cart.getTranslationalAcceleration()->y());
    EXPECT_DOUBLE_EQ(3.0, cart.getTranslationalAcceleration()->z());

    // Test orientation
    EXPECT_THROW(cart.getRotationalAcceleration(), std::runtime_error);
}
